__all__ = ()


def _merge_to_type(iterable_0, iterable_1, target_type):
    """
    Merges the two iterable to one of the given type.
    
    Parameters
    ----------
    iterable_0 : `None | iterable`
        the first iterable to merge.
    
    iterable_1 : `None | iterable`
        the second iterable to merge.
    
    target_type : `type`
        The expected output type.
    
    Returns
    -------
    merged : `None | instance<target_type>`
        Returns `None` if both `iterable_0` and `iterable_1` are `None`.
    """
    if iterable_0 is None:
        if iterable_1 is None:
            merged = None
        else:
            if type(iterable_1) is target_type:
                merged = iterable_1
            else:
                merged = target_type(iterable_1)
    
    else:
        if iterable_1 is None:
            if type(iterable_0) is target_type:
                merged = iterable_0
            else:
                merged = target_type(iterable_0)
        
        else:
            merged = [*iterable_0, *iterable_1]
            
            if target_type is not list:
                merged = target_type(merged)
    
    return merged



__all__ = ()

from ...application_command.application_command_permission.fields import validate_application_id, validate_guild_id
from ...application_command.application_command_permission_overwrite.fields import validate_channel_id

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_serializer_id


# ---- application_id ----

APPLICATION_ID_CONVERSION = AuditLogEntryDetailConversion(
    'application_id',
    'application_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_application_id,
)


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryDetailConversion(
    'channel_id',
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- guild_id ----

GUILD_ID_CONVERSION = AuditLogEntryDetailConversion(
    'guild_id',
    'guild_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_guild_id,
)


# ---- Construct ----

APPLICATION_COMMAND_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    APPLICATION_ID_CONVERSION,
    CHANNEL_ID_CONVERSION,
    GUILD_ID_CONVERSION,
)



__all__ = ()

from scarletio import RichAttributeErrorBaseType, docs_property


class PlaceHolderBase(RichAttributeErrorBaseType):
    __type_doc__ = (
    """
    Base type of place holders.
    
    Attributes
    ----------
    attribute_name : `None | str`
        The name of the place held attribute.
    
    docs : `None | str`
        Documentation of the place held attribute.
    
    type_name : `None | str`
        The name of the type the we place held at.
    """)
    
    @property
    def __instance_doc__(self):
        return self.docs
    
    __doc__ = docs_property()
    
    __slots__ = ('attribute_name', 'docs', 'type_name')

    def __new__(cls, docs = None):
        """
        Creates a new slot place holder.
        
        Parameters
        ----------
        docs : `None`, `str` = `None`, Optional
            Documentation of the place held attribute.
        """
        self = object.__new__(cls)
        self.attribute_name = None
        self.docs = docs
        self.type_name = None
        return self
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper function for ``.__eq__``. Returns whether the two instances are equal.
        Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # type_name
        if self.type_name != other.type_name:
            return False
        
        # docs
        if self.docs != other.docs:
            return False
        
        # attribute_name
        if self.attribute_name != other.attribute_name:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        type_name = self.type_name
        attribute_name = self.attribute_name
        if (type_name is None) or (attribute_name is None):
            field_added = False
        
        else:
            repr_parts.append(' of ')
            repr_parts.append(type_name)
            repr_parts.append('.')
            repr_parts.append(attribute_name)
            
            field_added = True
        
        field_added = self._put_repr_parts_into(repr_parts, field_added)
        
        docs = self.docs
        if (docs is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' docs = ')
            repr_parts.append(repr(docs))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts, field_added):
        """
        Helper function for sub-types to extend the created representation.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            Representation parts to extended.
        
        field_added : `bool`
            Whether a field was already added to representation.
        
        Returns
        -------
        field_added : `bool`
        """
        return field_added
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # docs
        docs = self.docs
        if (docs is not None):
            hash_value ^= hash(docs)
        
        # attribute_name
        attribute_name = self.attribute_name
        if (attribute_name is not None):
            hash_value ^= hash(attribute_name)
        
        # type_name
        type_name = self.type_name
        if (type_name is not None):
            hash_value ^= hash(type_name)
        
        return hash_value
    
    
    def __set_name__(self, owner, attribute_name):
        """
        Called when the type is constructed.
        
        Parameters
        ----------
        owner : `type`
            The parent type.
        
        attribute_name : `str`
            the name of the help attribute.
        """
        self.type_name = owner.__name__
        self.attribute_name = attribute_name
    
    
    def __get__(self, instance, type_):
        """
        Called when the place holder is accessed with a `get` operation.
        
        Returns
        -------
        self / default : `self` / `object`
            If accessed from class returns self. If from instance, returns the default object.
        """
        if instance is None:
            return self
        
        return NotImplemented
    
    
    def __set__(self, instance, value):
        """
        Called when place holder is accessed with a `set` operation.
        
        Raises
        ------
        NotImplementedError
        """
        if instance is None:
            type_name = self.type_name
        else:
            type_name = type(instance).__name__
        
        attribute_name = self.attribute_name
        if attribute_name is None:
            attribute_name = 'unknown'
        else:
            attribute_name = repr(attribute_name)
        
        raise NotImplementedError(
            f'Setting `{type_name}.{attribute_name}` is not supported; '
            f'got instance = {instance!r}; value = {value!r}.'
        )
    
    
    def __delete__(self, instance):
        """
        Called when place holder is accessed with a `del` operation.
        
        Raises
        ------
        NotImplementedError
        """
        if instance is None:
            type_name = self.type_name
        else:
            type_name = type(instance).__name__
        
        attribute_name = self.attribute_name
        if attribute_name is None:
            attribute_name = 'unknown'
        else:
            attribute_name = repr(attribute_name)
        
        raise NotImplementedError(
            f'Deleting `{type_name}.{attribute_name}` is not supported; '
            f'got instance = {instance!r}.'
        )



__all__ = ('create_partial_sticker_data', 'create_partial_sticker_from_id', 'create_partial_sticker_from_partial_data')


from .fields import put_format, put_id, put_name, parse_id, parse_format, parse_name
from .sticker import Sticker
from ...core import STICKERS


def create_partial_sticker_data(sticker):
    """
    Creates partial sticker data for the given sticker.
    
    Parameters
    ----------
    sticker : ``Sticker``
        The sticker to create the partial data from.
    
    Returns
    ------
    data : `dict` of (`str`, `object`) items
    """
    data = {}
    put_format(sticker.format, data, True)
    put_id(sticker.id, data, True)
    put_name(sticker.name, data, True)
    return data


def create_partial_sticker_from_partial_data(data):
    """
    Creates a sticker from the given partial sticker data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Sticker data.
    
    Returns
    -------
    sticker : ``Sticker``
    """
    sticker_id = parse_id(data)
    
    try:
        sticker = STICKERS[sticker_id]
    except KeyError:
        sticker = Sticker._create_empty(sticker_id)
        STICKERS[sticker_id] = sticker
    else:
        if not sticker.partial:
            return sticker
    
    sticker.format = parse_format(data)
    sticker.name = parse_name(data)
    
    return sticker


def create_partial_sticker_from_id(sticker_id):
    """
    Creates a sticker from the given identifier.
    
    Parameters
    ----------
    sticker_id : `int`
        The identifier to create the sticker for.
    
    Returns
    -------
    sticker : ``Sticker``
    """
    try:
        sticker = STICKERS[sticker_id]
    except KeyError:
        sticker = Sticker._create_empty(sticker_id)
        STICKERS[sticker_id] = sticker
    
    return sticker



__all__ = ()

from ...field_parsers import int_parser_factory, preinstanced_parser_factory
from ...field_putters import int_putter_factory, preinstanced_putter_factory
from ...field_validators import int_conditional_validator_factory, preinstanced_validator_factory

from .preinstanced import ScheduleWeeksDay


# nth_week

parse_nth_week = int_parser_factory('n', 1)
put_nth_week = int_putter_factory('n')
validate_nth_week = int_conditional_validator_factory(
    'nth_week',
    1,
    (lambda nth_week : nth_week >= 1 and nth_week <= 5),
    '>= 1 and <= 5',
)


# day

parse_weeks_day = preinstanced_parser_factory('day', ScheduleWeeksDay, ScheduleWeeksDay.monday)
put_weeks_day = preinstanced_putter_factory('day')
validate_weeks_day = preinstanced_validator_factory('weeks_day', ScheduleWeeksDay)



__all__ = ('AuditLogEntryType', 'AuditLogEntryTargetType')

from scarletio import copy_docs, export

from ...bases import Preinstance as P, PreinstancedBase

from ..audit_log_entry_change_conversions.application_command import (
    APPLICATION_COMMAND_CONVERSIONS as CHANGE_APPLICATION_COMMAND_CONVERSIONS
)
from ..audit_log_entry_change_conversions.auto_moderation_rule import (
    AUTO_MODERATION_RULE_CONVERSIONS as CHANGE_AUTO_MODERATION_RULE_CONVERSIONS
)
from ..audit_log_entry_change_conversions.channel import CHANNEL_CONVERSIONS as CHANGE_CHANNEL_CONVERSIONS
from ..audit_log_entry_change_conversions.channel_permission_overwrite import (
    CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS as CHANGE_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_change_conversions.emoji import EMOJI_CONVERSIONS as CHANGE_EMOJI_CONVERSIONS
from ..audit_log_entry_change_conversions.guild import GUILD_CONVERSIONS as CHANGE_GUILD_CONVERSIONS
from ..audit_log_entry_change_conversions.integration import INTEGRATION_CONVERSIONS as CHANGE_INTEGRATION_CONVERSIONS
from ..audit_log_entry_change_conversions.invite import INVITE_CONVERSIONS as CHANGE_INVITE_CONVERSIONS
from ..audit_log_entry_change_conversions.onboarding_prompt import (
    ONBOARDING_PROMPT_CONVERSIONS as CHANGE_ONBOARDING_PROMPT_CONVERSIONS
)
from ..audit_log_entry_change_conversions.onboarding_screen import (
    ONBOARDING_SCREEN_CONVERSIONS as CHANGE_ONBOARDING_SCREEN_CONVERSIONS
)
from ..audit_log_entry_change_conversions.role import ROLE_CONVERSIONS as CHANGE_ROLE_CONVERSIONS
from ..audit_log_entry_change_conversions.scheduled_event import (
    SCHEDULED_EVENT_CONVERSIONS as CHANGE_SCHEDULED_EVENT_CONVERSIONS
)
from ..audit_log_entry_change_conversions.soundboard_sound import (
    SOUNDBOARD_SOUND_CONVERSIONS as CHANGE_SOUNDBOARD_SOUND_CONVERSIONS
)
from ..audit_log_entry_change_conversions.stage import STAGE_CONVERSIONS as CHANGE_STAGE_CONVERSIONS
from ..audit_log_entry_change_conversions.sticker import STICKER_CONVERSIONS as CHANGE_STICKER_CONVERSIONS
from ..audit_log_entry_change_conversions.user import USER_CONVERSIONS as CHANGE_USER_CONVERSIONS
from ..audit_log_entry_change_conversions.webhook import WEBHOOK_CONVERSIONS as CHANGE_WEBHOOK_CONVERSIONS
from ..audit_log_entry_detail_conversions.application_command import (
    APPLICATION_COMMAND_CONVERSIONS as DETAIL_APPLICATION_COMMAND_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.auto_moderation_action_execution import (
    AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS as DETAIL_AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.channel import CHANNEL_CONVERSIONS as DETAIL_CHANNEL_CONVERSIONS
from ..audit_log_entry_detail_conversions.channel_permission_overwrite import (
    CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS as DETAIL_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS
)
from ..audit_log_entry_detail_conversions.guild import GUILD_CONVERSIONS as DETAIL_GUILD_CONVERSIONS
from ..audit_log_entry_detail_conversions.message import MESSAGE_CONVERSIONS as DETAIL_MESSAGE_CONVERSIONS
from ..audit_log_entry_detail_conversions.stage import STAGE_CONVERSIONS as DETAIL_STAGE_CONVERSIONS
from ..audit_log_entry_detail_conversions.user import USER_CONVERSIONS as DETAIL_USER_CONVERSIONS

from .target_converters import (
    target_converter_application_command, target_converter_auto_moderation_rule, target_converter_channel,
    target_converter_emoji, target_converter_guild, target_converter_integration, target_converter_invite,
    target_converter_role, target_converter_scheduled_event, target_converter_soundboard_sound, target_converter_stage,
    target_converter_sticker, target_converter_user, target_converter_webhook
)


class AuditLogEntryTargetType(PreinstancedBase, value_type = int):
    """
    Represents the target type of an ``AuditLogEntry``.
    
    Attributes
    ----------
    change_conversions : `None | AuditLogEntryDetailConversionGroup`
        Change conversions.
    
    detail_conversions : `None | AuditLogEntryDetailConversionGroup`
        Detail conversions.
    
    name : `str`
        The name of audit log target type.
    
    target_converter : `None | FunctionType | MethodType`
        Audit log target converter.
    
    value : `int`
        The identifier value of the audit log target type. Only used for hashing.
    
    Type Attributes
    ---------------
    Every predefined audit log target type can be accessed as type attribute as well:
    
    +-----------------------------------+-----------------------------------+-------+
    | Type attribute name               | name                              | value |
    +===================================+===================================+=======+
    | none                              | none                              | 0     |
    +-----------------------------------+-----------------------------------+-------+
    | guild                             | guild                             | 1     |
    +-----------------------------------+-----------------------------------+-------+
    | channel                           | channel                           | 2     |
    +-----------------------------------+-----------------------------------+-------+
    | channel_permission_overwrite      | channel permission overwrite      | 3     |
    +-----------------------------------+-----------------------------------+-------+
    | user                              | user                              | 4     |
    +-----------------------------------+-----------------------------------+-------+
    | role                              | role                              | 5     |
    +-----------------------------------+-----------------------------------+-------+
    | invite                            | invite                            | 6     |
    +-----------------------------------+-----------------------------------+-------+
    | webhook                           | webhook                           | 7     |
    +-----------------------------------+-----------------------------------+-------+
    | emoji                             | emoji                             | 8     |
    +-----------------------------------+-----------------------------------+-------+
    | integration                       | integration                       | 9     |
    +-----------------------------------+-----------------------------------+-------+
    | stage                             | stage                             | 10    |
    +-----------------------------------+-----------------------------------+-------+
    | scheduled_event                   | scheduled event                   | 11    |
    +-----------------------------------+-----------------------------------+-------+
    | sticker                           | sticker                           | 12    |
    +-----------------------------------+-----------------------------------+-------+
    | application_command               | application command               | 13    |
    +-----------------------------------+-----------------------------------+-------+
    | auto_moderation_rule              | auto moderation rule              | 14    |
    +-----------------------------------+-----------------------------------+-------+
    | auto_moderation_action_execution  | auto moderation action execution  | 15    |
    +-----------------------------------+-----------------------------------+-------+
    | onboarding_screen                 | onboarding screen                 | 16    |
    +-----------------------------------+-----------------------------------+-------+
    | onboarding_prompt                 | onboarding prompt                 | 17    |
    +-----------------------------------+-----------------------------------+-------+
    | message                           | message                           | 18    |
    +-----------------------------------+-----------------------------------+-------+
    | soundboard_sound                  | soundboard sound                  | 19    |
    +-----------------------------------+-----------------------------------+-------+
    """
    __slots__ = ('change_conversions', 'detail_conversions', 'target_converter')
    
    def __new__(cls, value, name = None, target_converter = None, change_conversions = None, detail_conversions = None):
        """
        Creates a new audit log target type.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the audit log target type.
        
        name : `None | str` = `None`, Optional
            The name of the audit log target type.
        
        target_converter : `None | FunctionType | MethodType` = `None`, Optional
            Audit log target converter.
        
        change_conversions :  `None | AuditLogEntryChangeConversionGroup` = `None`, Optional
            Change conversions.
        
        detail_conversions : `None | AuditLogEntryDetailConversionGroup` = `None`, Optional
            Detail conversions.
        """
        self = PreinstancedBase.__new__(cls, value, name)
        self.target_converter = target_converter
        self.change_conversions = change_conversions
        self.detail_conversions = detail_conversions
        return self
    

    # predefined
    none = P(
        0,
        'none',
        None,
        None,
        None,
    )
    guild = P(
        1,
        'guild',
        target_converter_guild,
        CHANGE_GUILD_CONVERSIONS,
        DETAIL_GUILD_CONVERSIONS,
    )
    channel = P(
        2,
        'channel',
        target_converter_channel,
        CHANGE_CHANNEL_CONVERSIONS,
        DETAIL_CHANNEL_CONVERSIONS,
    )
    channel_permission_overwrite = P(
        3,
        'channel permission overwrite',
        target_converter_channel,
        CHANGE_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS,
        DETAIL_CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS,
    )
    user = P(
        4,
        'user',
        target_converter_user,
        CHANGE_USER_CONVERSIONS,
        DETAIL_USER_CONVERSIONS,
    )
    role = P(
        5,
        'role',
        target_converter_role,
        CHANGE_ROLE_CONVERSIONS,
        None,
    )
    invite = P(
        6,
        'invite',
        target_converter_invite,
        CHANGE_INVITE_CONVERSIONS,
        None,
    )
    webhook = P(
        7,
        'webhook',
        target_converter_webhook,
        CHANGE_WEBHOOK_CONVERSIONS,
        None,
    )
    emoji = P(
        8,
        'emoji',
        target_converter_emoji,
        CHANGE_EMOJI_CONVERSIONS,
        None,
    )
    integration = P(
        9,
        'integration',
        target_converter_integration,
        CHANGE_INTEGRATION_CONVERSIONS,
        None,
    )
    stage = P(
        10,
        'stage',
        target_converter_stage,
        CHANGE_STAGE_CONVERSIONS,
        DETAIL_STAGE_CONVERSIONS,
    )
    scheduled_event = P(
        11,
        'scheduled event',
        target_converter_scheduled_event,
        CHANGE_SCHEDULED_EVENT_CONVERSIONS,
        None,
    )
    sticker = P(
        12,
        'sticker',
        target_converter_sticker,
        CHANGE_STICKER_CONVERSIONS,
        None,
    )
    application_command = P(
        13,
        'application command',
        target_converter_application_command,
        CHANGE_APPLICATION_COMMAND_CONVERSIONS,
        DETAIL_APPLICATION_COMMAND_CONVERSIONS,
    )
    auto_moderation_rule = P(
        14,
        'auto moderation rule',
        target_converter_auto_moderation_rule,
        CHANGE_AUTO_MODERATION_RULE_CONVERSIONS,
        None,
    )
    auto_moderation_action_execution = P(
        15,
        'auto moderation action execution',
        None,
        CHANGE_USER_CONVERSIONS,
        DETAIL_AUTO_MODERATION_ACTION_EXECUTION_CONVERSIONS,
    )
    onboarding_screen = P(
        16,
        'onboarding screen',
        None,
        CHANGE_ONBOARDING_SCREEN_CONVERSIONS,
        None,
    )
    onboarding_prompt = P(
        17,
        'onboarding prompt',
        None,
        CHANGE_ONBOARDING_PROMPT_CONVERSIONS,
        None,
    )
    message = P(
        18,
        'message',
        None,
        None,
        DETAIL_MESSAGE_CONVERSIONS,
    )
    soundboard_sound = P(
        19,
        'soundboard sound',
        target_converter_soundboard_sound,
        CHANGE_SOUNDBOARD_SOUND_CONVERSIONS,
        None,
    )


@export
class AuditLogEntryType(PreinstancedBase, value_type = int):
    """
    Represents the event type of an ``AuditLogEntry``.
    
    Attributes
    ----------
    name : `str`
        The name of audit log event.
    
    target_type : ``AuditLogEntryTargetType``
        The audit log's target's type.
    
    value : `int`
        The Discord side identifier value of the audit log event.
    
    Type Attributes
    ---------------
    Every predefined audit log event can be accessed as type attribute as well:
    
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | Type attribute name                   | Name                                  | Value | Target type                       |
    +=======================================+=======================================+=======+===================================+
    | none                                  | none                                  |  0    | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | guild_update                          | guild update                          |  1    | guild                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_create                        | channel create                        | 10    | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_update                        | channel update                        | 11    | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_delete                        | channel delete                        | 12    | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_permission_overwrite_create   | channel permission overwrite create   | 13    | channel_permission_overwrite      |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_permission_overwrite_update   | channel permission overwrite update   | 14    | channel_permission_overwrite      |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_permission_overwrite_delete   | channel permission overwrite delete   | 15    | channel_permission_overwrite      |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_kick                             | user kick                             | 20    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_prune                            | user prune                            | 21    | guild                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_ban_add                          | user ban add                          | 22    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_ban_remove                       | user ban remove                       | 23    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_update                           | user update                           | 24    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_role_update                      | user role update                      | 25    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_move                             | user move                             | 26    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | user_disconnect                       | user disconnect                       | 27    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | bot_add                               | bot add                               | 28    | user                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_create                           | role create                           | 30    | role                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_update                           | role update                           | 31    | role                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_delete                           | role delete                           | 32    | role                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | invite_create                         | invite create                         | 40    | invite                            |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | invite_update                         | invite update                         | 41    | invite                            |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | invite_delete                         | invite delete                         | 42    | invite                            |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | webhook_create                        | webhook create                        | 50    | webhook                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | webhook_update                        | webhook update                        | 51    | webhook                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | webhook_delete                        | webhook delete                        | 52    | webhook                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | emoji_create                          | emoji create                          | 60    | emoji                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | emoji_update                          | emoji update                          | 61    | emoji                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | emoji_delete                          | emoji delete                          | 62    | emoji                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | message_delete                        | message delete                        | 72    | message                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | message_bulk_delete                   | message bulk delete                   | 73    | message                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | message_pin                           | message pin                           | 74    | message                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | message_unpin                         | message unpin                         | 75    | message                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | integration_create                    | integration create                    | 80    | integration                       |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | integration_update                    | integration update                    | 81    | integration                       |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | integration_delete                    | integration delete                    | 82    | integration                       |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | stage_create                          | stage create                          | 83    | stage                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | stage_update                          | stage update                          | 84    | stage                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | stage_delete                          | stage delete                          | 85    | stage                             |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | sticker_create                        | sticker create                        | 90    | sticker                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | sticker_update                        | sticker update                        | 91    | sticker                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | sticker_delete                        | sticker delete                        | 92    | sticker                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | scheduled_event_create                | scheduled event create                | 100   | scheduled_event                   |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | scheduled_event_update                | scheduled event update                | 101   | scheduled_event                   |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | scheduled_event_delete                | scheduled event delete                | 102   | scheduled_event                   |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | thread_create                         | thread create                         | 110   | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | thread_update                         | thread update                         | 111   | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | thread_delete                         | thread delete                         | 112   | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | application_command_permission_update | application command permission update | 121   | application_command               |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | soundboard_sound_create               | soundboard sound create               | 130   | soundboard_sound                  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | soundboard_sound_update               | soundboard sound update               | 131   | soundboard_sound                  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | soundboard_sound_delete               | soundboard sound delete               | 132   | soundboard_sound                  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_rule_create           | auto moderation rule create           | 140   | auto_moderation_rule              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_rule_update           | auto moderation rule update           | 141   | auto_moderation_rule              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_rule_delete           | auto moderation rule delete           | 142   | auto_moderation_rule              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_block_message         | auto moderation block message         | 143   | auto_moderation_action_execution  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_alert_message         | auto moderation alert message         | 144   | auto_moderation_action_execution  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | auto_moderation_user_timeout          | auto moderation user timeout          | 145   | auto_moderation_action_execution  |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | creator_monetization_request_created  | creator_monetization_request_created  | 150   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | creator_monetization_terms_accepted   | creator_monetization_terms_accepted   | 151   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_prompt_create                    | role_prompt_create                    | 160   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_prompt_update                    | role_prompt_update                    | 161   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | role_prompt_delete                    | role_prompt_delete                    | 162   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | onboarding_prompt_create              | onboarding prompt create              | 163   | onboarding_prompt                 |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | onboarding_prompt_update              | onboarding prompt update              | 164   | onboarding_prompt                 |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | onboarding_prompt_delete              | onboarding prompt delete              | 165   | onboarding_prompt                 |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | onboarding_screen_create              | onboarding screen create              | 166   | onboarding_screen                 |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | onboarding_screen_update              | onboarding screen update              | 167   | onboarding_screen                 |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | home_feature_item                     | home feature item                     | 171   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | home_remove_item                      | home remove item                      | 172   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | home_screen_create                    | home screen create                    | 190   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | home_screen_update                    | home screen update                    | 191   | none                              |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_status_update                 | channel status update                 | 192   | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    | channel_status_delete                 | channel status delete                 | 193   | channel                           |
    +---------------------------------------+---------------------------------------+-------+-----------------------------------+
    """
    __slots__ = ('target_type',)
    
    def __new__(cls, value, name = None, target_type = None):
        """
        Creates a new audit log event.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the audit log event.
        
        name : `None | str` = `None`, Optional
            The audit log event's name.
        
        target_type : `None | AuditLogEntryTargetType` = `None`, Optional
            The audit log's target type.
        """
        if target_type is None:
            target_type = AuditLogEntryTargetType.none
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.target_type = target_type
        return self
    
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new audit log event object from the given value.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value what has no representation yet.
        
        Returns
        -------
        self : ``AuditLogEntryType``
            The created audit log event.
        """
        self = super(AuditLogEntryType, cls)._from_value(value)
        self.target_type = AuditLogEntryTargetType.none
        return self
    
    
    @copy_docs(PreinstancedBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(', target_type = ')
        repr_parts.append(self.target_type.name)
    
    
    # predefined
    none = P(0, 'none', AuditLogEntryTargetType.none)
    
    guild_update = P(1, 'guild update', AuditLogEntryTargetType.guild)
    
    channel_create = P(10, 'channel create', AuditLogEntryTargetType.channel)
    channel_update = P(11, 'channel update', AuditLogEntryTargetType.channel)
    channel_delete = P(12, 'channel delete', AuditLogEntryTargetType.channel)
    
    channel_permission_overwrite_create = P(
        13,
        'channel permission overwrite create',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    channel_permission_overwrite_update = P(
        14,
        'channel permission overwrite update',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    channel_permission_overwrite_delete = P(
        15,
        'channel permission overwrite delete',
        AuditLogEntryTargetType.channel_permission_overwrite,
    )
    
    user_kick = P(20, 'user kick', AuditLogEntryTargetType.user)
    user_prune = P(21, 'user prune', AuditLogEntryTargetType.guild)
    user_ban_add = P(22, 'user ban add', AuditLogEntryTargetType.user)
    user_ban_remove = P(23, 'user ban remove', AuditLogEntryTargetType.user)
    user_update = P(24, 'user update', AuditLogEntryTargetType.user)
    user_role_update = P(25, 'user role update', AuditLogEntryTargetType.user)
    user_move = P(26, 'user move', AuditLogEntryTargetType.user)
    user_disconnect = P(27, 'user disconnect', AuditLogEntryTargetType.user)
    bot_add = P(28, 'user role update', AuditLogEntryTargetType.user)
    
    role_create = P(30, 'role create', AuditLogEntryTargetType.role)
    role_update = P(31, 'role update', AuditLogEntryTargetType.role)
    role_delete = P(32, 'role delete', AuditLogEntryTargetType.role)
    
    invite_create = P(40, 'invite create', AuditLogEntryTargetType.invite)
    invite_update = P(41, 'invite update', AuditLogEntryTargetType.invite)
    invite_delete = P(42, 'invite delete', AuditLogEntryTargetType.invite)
    
    webhook_create = P(50, 'webhook create', AuditLogEntryTargetType.webhook)
    webhook_update = P(51, 'webhook update', AuditLogEntryTargetType.webhook)
    webhook_delete = P(52, 'webhook delete', AuditLogEntryTargetType.webhook)
    
    emoji_create = P(60, 'emoji create', AuditLogEntryTargetType.emoji)
    emoji_update = P(61, 'emoji update', AuditLogEntryTargetType.emoji)
    emoji_delete = P(62, 'emoji delete', AuditLogEntryTargetType.emoji)
    
    message_delete = P(72, 'message delete', AuditLogEntryTargetType.message)
    message_bulk_delete = P(73, 'message bulk delete', AuditLogEntryTargetType.message)
    message_pin = P(74, 'message pin', AuditLogEntryTargetType.message)
    message_unpin = P(75, 'message unpin', AuditLogEntryTargetType.message)
    
    integration_create = P(80, 'integration create', AuditLogEntryTargetType.integration)
    integration_update = P(81, 'integration update', AuditLogEntryTargetType.integration)
    integration_delete = P(82, 'integration delete', AuditLogEntryTargetType.integration)
    
    stage_create = P(83, 'stage create', AuditLogEntryTargetType.stage)
    stage_update = P(84, 'stage update', AuditLogEntryTargetType.stage)
    stage_delete = P(85, 'stage delete', AuditLogEntryTargetType.stage)
    
    sticker_create = P(90, 'sticker create', AuditLogEntryTargetType.sticker)
    sticker_update = P(91, 'sticker update', AuditLogEntryTargetType.sticker)
    sticker_delete = P(92, 'sticker delete', AuditLogEntryTargetType.sticker)
    
    scheduled_event_create = P(100, 'scheduled event create', AuditLogEntryTargetType.scheduled_event)
    scheduled_event_update = P(101, 'scheduled event update', AuditLogEntryTargetType.scheduled_event)
    scheduled_event_delete = P(102, 'scheduled event delete', AuditLogEntryTargetType.scheduled_event)
    
    thread_create = P(110, 'thread create', AuditLogEntryTargetType.channel)
    thread_update = P(111, 'thread update', AuditLogEntryTargetType.channel)
    thread_delete = P(112, 'thread delete', AuditLogEntryTargetType.channel)
    
    application_command_permission_update = P(
        121,
        'application command permission update',
        AuditLogEntryTargetType.application_command,
    )
    
    soundboard_sound_create =  P(130,'soundboard sound create', AuditLogEntryTargetType.soundboard_sound)
    soundboard_sound_update =  P(131, 'soundboard sound update', AuditLogEntryTargetType.soundboard_sound)
    soundboard_sound_delete = P(132, 'soundboard sound delete', AuditLogEntryTargetType.soundboard_sound)
    
    auto_moderation_rule_create = P(140, 'auto moderation rule create', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_rule_update = P(141, 'auto moderation rule update', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_rule_delete = P(142, 'auto moderation rule delete', AuditLogEntryTargetType.auto_moderation_rule)
    auto_moderation_block_message = P(
        143, 'auto moderation block message', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    auto_moderation_alert_message = P(
        144, 'auto moderation alert message', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    auto_moderation_user_timeout = P(
        145, 'auto moderation user timeout', AuditLogEntryTargetType.auto_moderation_action_execution
    )
    
    creator_monetization_request_created = P(150, 'creator monetization request created', AuditLogEntryTargetType.none)
    creator_monetization_terms_accepted = P(151, 'creator monetization terms accepted', AuditLogEntryTargetType.none)
    
    # These 3 are deprecated
    role_prompt_create =  P(160,'role prompt create', AuditLogEntryTargetType.none)
    role_prompt_update =  P(161, 'role prompt update', AuditLogEntryTargetType.none)
    role_prompt_delete = P(162, 'role prompt delete', AuditLogEntryTargetType.none)
    
    onboarding_prompt_create =  P(163,'onboarding prompt create', AuditLogEntryTargetType.onboarding_prompt)
    onboarding_prompt_update =  P(164, 'onboarding prompt update', AuditLogEntryTargetType.onboarding_prompt)
    onboarding_prompt_delete = P(165, 'onboarding prompt delete', AuditLogEntryTargetType.onboarding_prompt)
    
    onboarding_screen_create =  P(166,'onboarding screen create', AuditLogEntryTargetType.onboarding_screen)
    onboarding_screen_update =  P(167, 'onboarding screen update', AuditLogEntryTargetType.onboarding_screen)
    
    home_feature_item = P(171, 'home feature item', AuditLogEntryTargetType.none)
    home_remove_item = P(172, 'home remove item', AuditLogEntryTargetType.none)
    
    home_screen_create = P(190, 'home screen create', AuditLogEntryTargetType.none)
    home_screen_update = P(191, 'home screen update', AuditLogEntryTargetType.none)
    channel_status_update = P(192, 'channel status update', AuditLogEntryTargetType.channel)
    channel_status_delete = P(193, 'channel status delete', AuditLogEntryTargetType.channel)



__all__ = ()

from scarletio import Compound, set_docs
from scarletio.web_common import FormData

from ....env import API_VERSION

from ...bases import maybe_snowflake, maybe_snowflake_pair
from ...builder.serialization import create_serializer
from ...builder.serialization_configuration import SerializationConfiguration
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...channel.channel.utils import CHANNEL_GUILD_THREAD_FIELD_CONVERTERS
from ...core import CHANNELS
from ...http import DiscordApiClient
from ...message import Message
from ...message.message_builder import MessageBuilderForumThreadCreate
from ...payload_building import add_payload_fields_from_keyword_parameters, build_create_payload
from ...user import ClientUserBase, create_partial_user_from_id, create_user_from_thread_user_data, thread_user_create

from ..functionality_helpers import request_channel_thread_channels
from ..request_helpers import (
    get_channel_and_id, get_channel_id, get_guild_id, get_channel_guild_id_and_id, get_user_id,
)


MESSAGE_SERIALIZER_FORUM_THREAD_CREATE = create_serializer(
    MessageBuilderForumThreadCreate,
    SerializationConfiguration(
        [
            MessageBuilderForumThreadCreate.allowed_mentions,
            MessageBuilderForumThreadCreate.attachments,
            MessageBuilderForumThreadCreate.components,
            MessageBuilderForumThreadCreate.content,
            MessageBuilderForumThreadCreate.embeds,
            MessageBuilderForumThreadCreate.flags,
            MessageBuilderForumThreadCreate.nonce,
            MessageBuilderForumThreadCreate.sticker_ids,
            MessageBuilderForumThreadCreate.tts,
        ],
        False,
    )
)


class ClientCompoundThreadEndpoints(Compound):
    
    api : DiscordApiClient
    id: int
    
    
    async def guild_thread_get_all_active(self, guild):
        """
        Gets all the active threads of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to get it's threads of.
            
            If the guild is given as `0`, will return an empty list.
        
        Returns
        -------
        threads : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `guild` is neither ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        if not guild_id:
            return []
        
        data = await self.api.guild_thread_get_all_active(guild_id)
        
        thread_channel_datas = data['threads']
        
        thread_channels = [
            Channel.from_data(thread_channel_data, self, guild_id) for thread_channel_data in thread_channel_datas
        ]
        
        thread_user_datas = data['members']
        for thread_user_data in thread_user_datas:
            thread_channel_id = int(thread_user_data['id'])
            try:
                thread_channel = CHANNELS[thread_channel_id]
            except KeyError:
                continue
    
            user_id = int(thread_user_data['user_id'])
            user = create_partial_user_from_id(user_id)
            
            thread_user_create(thread_channel, user, thread_user_data)
        
        return thread_channels
    
    
    async def thread_create(self, message_or_channel, channel_template = None, **keyword_parameters,):
        """
        Creates a new thread derived from the given message or channel.
        
        > For private thread channels the guild needs to have level 2 boost level.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message_or_channel : ``Channel``, ``Message``, `int`, `tuple` (`int`, `int`)
            The channel or message to create thread from.
            
            > If given as a channel instance, will create a private thread, else a public one.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            Channel to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        applied_tag_ids : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
             The tags' identifier which have been applied to the thread.
        
        applied_tags : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
            Alternative for `applied_tag_ids`.
        
        auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        channel_type : `None`, ``ChannelType``, `int` = `None`, Optional (Keyword only)
            The type of the created (thread) channel.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        open_ : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        Returns
        -------
        thread_channel : ``Channel``
            The created thread channel.
        
        Raises
        ------
        TypeError
            If any parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # Message check order
        # 1.: Message
        # 4.: None -> raise
        # 5.: `tuple` (`int`, `int`)
        # 6.: raise
        #
        # Message cannot be detected by id, only cached ones, so ignore that case.
        
        if isinstance(message_or_channel, Channel):
            if not message_or_channel.is_threadable():
                raise TypeError(
                    f'{message_or_channel!r} do not supports thread creation.'
                )
            
            message_id = None
            channel = message_or_channel
            channel_id = channel.id
        
        elif isinstance(message_or_channel, Message):
            message_id = message_or_channel.id
            channel_id = message_or_channel.channel_id
            channel = CHANNELS.get(channel_id, None)
        
        else:
            channel_id = maybe_snowflake(message_or_channel)
            if (channel_id is not None):
                message_id = None
                channel = CHANNELS.get(channel_id, None)
            
            else:
                snowflake_pair = maybe_snowflake_pair(message_or_channel)
                if snowflake_pair is None:
                    raise TypeError(
                        f'`message_or_channel` can be `{Channel.__name__}`, `{Message.__name__}`, '
                        f'`int`, `tuple` (`int`, `int`), '
                        f'got {message_or_channel.__class__.__name__}; {message_or_channel!r}.'
                    )
                
                channel_id, message_id = snowflake_pair
                channel = CHANNELS.get(channel_id)
                # Checkout type
        
        if message_id is None:
            channel_type = ChannelType.guild_thread_private
        else:
            channel_type = ChannelType.guild_thread_public
        keyword_parameters.setdefault('channel_type', channel_type)
        
        data = build_create_payload(channel_template, CHANNEL_GUILD_THREAD_FIELD_CONVERTERS, keyword_parameters)
        
        if message_id is None:
            coroutine = self.api.thread_create(channel_id, data)
        else:
            coroutine = self.api.thread_create_from_message(channel_id, message_id, data)
        channel_data = await coroutine
        
        if channel is None:
            guild_id = channel_data.get('guild_id', None)
            if (guild_id is not None):
                guild_id = int(guild_id)
        else:
            guild_id = channel.guild_id
        
        return Channel.from_data(channel_data, self, guild_id)
    
    
    async def forum_thread_create(
        self, channel_forum, channel_template = None, *positional_parameters, **keyword_parameters,
    ):
        """
        Creates and thread at the given `channel` with the given message fields. If there is nothing to send will
        return `None` for channel and `None` for message as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel_forum : ``Channel``, `int`
            The forum channel's identifier where the thread will be started.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            (Thread) channel entity to use as a template.
        
        *positional_parameters : Positional parameters
            Additional parameters to create the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        applied_tag_ids : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
             The tags' identifier which have been applied to the thread.
        
        applied_tags : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
            Alternative for `applied_tag_ids`.
        
        attachments : `None`, `object`, Optional (Keyword only)
            Attachments to send.
        
        auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
        
        content : `None`, `str`, Optional
            The message's content if given.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
        
        file : `None`, `object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None`, `object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags. Due to name collision, `message.flags` is not directly supported.
        
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        nonce : `None`, `str`, Optional (Keyword only)
            Used for optimistic message sending.
        
        open_ : `bool`, Optional (Keyword only)
            Whether the thread is open.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        sticker : `None`, ``Sticker``, Optional
            Alternative for `sticker_ids`.
        
        sticker_ids : `None`, `list<int>`, Optional (Keyword only)
            Sticker(s) to send within the message.
        
        stickers : `None`, `list<int | Sticker>`, Optional
            Alternative for `sticker_ids`.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        Returns
        -------
        thread_channel : `None`, ``Channel``
            The created thread channel. `None` if there was nothing to send.
        message : `None`, ``Message``
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.message_create`` : Sending a message to a text channel.
        ``.thread_create`` : Create thread in a text channel.
        """
        channel, channel_id = get_channel_and_id(channel_forum, Channel.is_in_group_forum)
        
        # channel_data
        if (channel_template is None):
            data = {}
        else:
            data = channel_template.to_data(defaults = False)
        
        keyword_parameters = add_payload_fields_from_keyword_parameters(
            CHANNEL_GUILD_THREAD_FIELD_CONVERTERS, keyword_parameters, data, False, raise_unused = False,
        )
        
        if keyword_parameters is None:
            keyword_parameters = {}
        
        message_data = MESSAGE_SERIALIZER_FORUM_THREAD_CREATE(positional_parameters, keyword_parameters)
        if not message_data:
            return None, None
        
        # Nest `message_data` under `data['message']`
        if isinstance(message_data, FormData):
            field = message_data.fields[0]
            data['message'] = field.value
            field.value = data
            data = message_data
        else:
            data['message'] = message_data
        
        channel_data = await self.api.thread_create(channel_id, data)
        
        if channel is None:
            guild_id = channel_data.get('guild_id', None)
            if (guild_id is None):
                guild_id = 0
            else:
                guild_id = int(guild_id)
        else:
            guild_id = channel.guild_id
        
        thread_channel = Channel.from_data(channel_data, self, guild_id)
        message_data = channel_data.get('message', None)
        if (message_data is None):
            message = None
        else:
            message = thread_channel._create_new_message(message_data)
        
        return thread_channel, message
    
    
    async def thread_join(self, thread_channel):
        """
        Joins the client to the given thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to join to, or it's identifier.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        
        await self.api.thread_join(channel_id)
    
    
    async def thread_leave(self, thread_channel):
        """
        Leaves the client to the given thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to join to, or it's identifier.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        
        await self.api.thread_leave(channel_id)
    
    
    async def thread_user_get(self, thread_channel, user):
        """
        Gets a user's thread profile inside of a thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to get the user's thread profile of.
        user : ``ClientUserBase``, `int`
            The user to get it's thread profile of.
        
        Returns
        -------
        user : ``ClientUserBase``
            The user, who's thread profile was requested.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        thread_channel, channel_id = get_channel_and_id(thread_channel, Channel.is_in_group_thread)
        user_id = get_user_id(user)
        
        thread_user_data = await self.api.thread_user_get(channel_id, user_id, {'with_member': True})
        
        if thread_channel is None:
            thread_channel = create_partial_channel_from_id(channel_id, ChannelType.guild_thread_public, 0)
        
        user = create_user_from_thread_user_data(thread_channel, thread_user_data)
        thread_user_create(thread_channel, user, thread_user_data)
        
        return user
    
    
    async def thread_user_add(self, thread_channel, user):
        """
        Adds the user to the thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to add the user to, or it's identifier.
        user : ``ClientUserBase``, `int`
            The user to add to the the thread.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            coroutine = self.api.thread_join(channel_id)
        else:
            coroutine = self.api.thread_user_add(channel_id, user_id)
        await coroutine
    
    
    async def thread_user_delete(self, thread_channel, user):
        """
        Deletes the user to the thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to remove the user from, or it's identifier.
        user : ``ClientUserBase``, `int`
            The user to remove from the thread.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            coroutine = self.api.thread_leave(channel_id)
        else:
            coroutine = self.api.thread_user_delete(channel_id, user_id)
        await coroutine
    
    
    async def thread_user_get_all(self, thread_channel):
        """
        Gets all the users of a thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to get it's users of.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
            The created users.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        thread_channel, channel_id = get_channel_and_id(thread_channel, Channel.is_in_group_thread)
        
        data = {
            'after': 0,
            'limit': 100,
            'with_member': True,
        }
        
        users = []
        
        while True:
            thread_user_datas = await self.api.thread_user_get_chunk(channel_id, data)
            
            if thread_channel is None:
                thread_channel = create_partial_channel_from_id(channel_id, ChannelType.guild_thread_public, 0)
            
            for thread_user_data in thread_user_datas:
                user = create_user_from_thread_user_data(thread_channel, thread_user_data)
                thread_user_create(thread_channel, user, thread_user_data)
                users.append(user)
            
            if len(thread_user_datas) < 100:
                break
            
            data['after'] = users[-1].id
            continue
        
        return users
    
    
    if API_VERSION >= 10:
        async def channel_thread_get_all_active(self, channel):
            guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
            thread_channels = await self.guild_thread_get_all_active(guild_id)
            return [thread_channel for thread_channel in thread_channels if thread_channel.parent_id == channel_id]
    
    else:
        async def channel_thread_get_all_active(self, channel):
            guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
            return await request_channel_thread_channels(
                self, guild_id, channel_id, type(self.api).channel_thread_get_chunk_active,
            )
    
    set_docs(
        channel_thread_get_all_active,
        """
        Requests all the active threads of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        When using API v10 or later, this endpoint filters from ``.guild_thread_get_all_active`` method's return.
        Consider using that instead.
        
        Active threads can also be extracted from ``Guild.threads``.
        """
    )
    
    async def channel_thread_get_all_archived_private(self, channel):
        """
        Requests all the archived private of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.api).channel_thread_get_chunk_archived_private,
        )
    
    
    async def channel_thread_get_all_archived_public(self, channel):
        """
        Requests all the archived public threads of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.api).channel_thread_get_chunk_archived_public,
        )
    
    
    async def channel_thread_get_all_self_archived(self, channel):
        """
        Requests all the archived private threads by the client.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.api).channel_thread_get_chunk_self_archived,
    )



__all__ = ()

from ...component import ComponentType
from ...component.shared_constants import CUSTOM_ID_LENGTH_MAX
from ...field_parsers import (
    field_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    nullable_field_optional_putter_factory, nullable_object_array_optional_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ...field_validators import (
    field_validator_factory, nullable_object_array_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)

# components

parse_components = nullable_object_array_parser_factory('components', NotImplemented, include = 'InteractionComponent')
put_components = nullable_object_array_optional_putter_factory('components')
validate_components = nullable_object_array_validator_factory(
    'components', NotImplemented, include = 'InteractionComponent'
)

# custom_id

parse_custom_id = nullable_string_parser_factory('custom_id')
put_custom_id = url_optional_putter_factory('custom_id')
validate_custom_id = nullable_string_validator_factory('custom_id', 0, CUSTOM_ID_LENGTH_MAX)

# type

parse_type = preinstanced_parser_factory('type', ComponentType, ComponentType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ComponentType)

# value

parse_value = field_parser_factory('value')
put_value = nullable_field_optional_putter_factory('value')
validate_value = field_validator_factory('value')



__all__ = ('TrackEndEvent', 'TrackExceptionEvent', 'TrackStartEvent', 'TrackStuckEvent', 'PlayerWebSocketClosedEvent')

import reprlib

from scarletio import copy_docs

from ...discord.bases import EventBase

from .constants import (
    LAVALINK_KEY_END_REASON, LAVALINK_KEY_EXCEPTION_REASON_DEPRECATED, LAVALINK_KEY_EXCEPTION_REASON_OBJECT,
    LAVALINK_KEY_EXCEPTION_REASON_OBJECT_MESSAGE, LAVALINK_KEY_EXCEPTION_REASON_OBJECT_SEVERITY,
    LAVALINK_KEY_THRESHOLD_MS, LAVALINK_KEY_WEBSOCKET_CLOSE_BY_REMOTE, LAVALINK_KEY_WEBSOCKET_CLOSE_CODE,
    LAVALINK_KEY_WEBSOCKET_CLOSE_REASON
)


class TrackStartEvent(EventBase):
    """
    Represents an event when the track is started.
    
    Attributes
    ----------
    player : ``SolarPlayerBase``
        The player associated with the event.
    track : ``ConfiguredTrack``
        The started track.
    """
    __slots__ = ('player', 'track', )
    
    def __new__(cls, player, data):
        """
        Creates a new ``TrackStartEvent`` event from the given parameters.
        
        Parameters
        ----------
        player : ``SolarPlayerBase``
            The player associated with the event.
        data : `dict` of (`str`, `object`) items
            Event payload.
        """
        track = player.get_current()
        
        self = object.__new__(cls)
        self.player = player
        self.track = track
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = []
        
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        repr_parts.append(' player = ')
        repr_parts.append(repr(self.player))
        
        repr_parts.append(', track = ')
        repr_parts.append(repr(self.track))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.player
        yield self.track
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.player != other.player:
            return False
        
        if self.track != other.track:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        return hash(self.player) ^ hash(self.track)


class TrackEndEvent(EventBase):
    """
    Represents an event when the track is ended.
    
    Attributes
    ----------
    player : ``SolarPlayerBase``
        The player associated with the event.
    reason : `None`, `str`
        The reason why the track ended.
    track : ``ConfiguredTrack``
        The started track.
    """
    __slots__ = ('player', 'reason', 'track', )
    
    def __new__(cls, player, data):
        """
        Creates a new ``TrackEndEvent`` event from the given parameters.
        
        Parameters
        ----------
        player : ``SolarPlayerBase``
            The player associated with the event.
        data : `dict` of (`str`, `object`) items
            Event payload.
        """
        track = player.get_current()
        reason = data[LAVALINK_KEY_END_REASON]
        
        self = object.__new__(cls)
        self.player = player
        self.track = track
        self.reason = reason
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = []
        
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        repr_parts.append(' player = ')
        repr_parts.append(repr(self.player))
        
        repr_parts.append(', track = ')
        repr_parts.append(repr(self.track))
        
        reason = self.reason
        if (reason is not None):
            repr_parts.append(', reason = ')
            repr_parts.append(reprlib.repr(reason))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.player
        yield self.track
        yield self.reason
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.player != other.player:
            return False
        
        if self.reason != other.reason:
            return False
        
        if self.track != other.track:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # player
        hash_value ^= hash(self.player)
        
        # reason
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        # track
        hash_value ^= hash(self.track)
        
        return hash_value


class TrackStuckEvent(EventBase):
    """
    Represents an event when the currently playing track is stuck.
    
    It usually is not a lavalink issue, but an issue with the stream.
    
    Attributes
    ----------
    player : ``SolarPlayerBase``
        The player associated with the event.
    threshold : `float`
        The amount of time till the time is stuck in seconds.
    track : ``ConfiguredTrack``
        The stucking track.
    """
    __slots__ = ('player', 'threshold', 'track', )
    
    def __new__(cls, player, data):
        """
        Creates a new ``TrackStuckEvent`` event from the given parameters.
        
        Parameters
        ----------
        player : ``SolarPlayerBase``
            The player associated with the event.
        data : `dict` of (`str`, `object`) items
            Event payload.
        """
        track = player.get_current()
        threshold = data[LAVALINK_KEY_THRESHOLD_MS]*1000.0
        
        self = object.__new__(cls)
        self.player = player
        self.track = track
        self.threshold = threshold
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = []
        
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        repr_parts.append(' player = ')
        repr_parts.append(repr(self.player))
        
        repr_parts.append(', track = ')
        repr_parts.append(repr(self.track))
        
        repr_parts.append(', threshold = ')
        repr_parts.append(self.threshold.__format__('.3f'))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.player
        yield self.track
        yield self.threshold
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.player != other.player:
            return False
        
        if self.threshold != other.threshold:
            return False
        
        if self.track != other.track:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # player
        hash_value ^= hash(self.player)
        
        # threshold
        hash_value ^= hash(self.threshold)
        
        # track
        hash_value ^= hash(self.track)
        
        return hash_value


class TrackExceptionEvent(EventBase):
    """
    Represents an event when exception occurs when playing a track.
    
    It usually is not a lavalink issue, but an issue with the stream.
    
    Attributes
    ----------
    player : ``SolarPlayerBase``
        The player associated with the event.
    reason : `str`
        Error reason.
    severity : `str`
        The exception's severity.
    track : ``ConfiguredTrack``
        The stucking track.
    """
    __slots__ = ('player', 'reason', 'severity', 'track')
    
    def __new__(cls, player, data):
        """
        Creates a new ``TrackExceptionEvent`` event from the given parameters.
        
        Parameters
        ----------
        player : ``SolarPlayerBase``
            The player associated with the event.
        data : `dict` of (`str`, `object`) items
            Event payload.
        """
        try:
            reason_object = data[LAVALINK_KEY_EXCEPTION_REASON_OBJECT]
        except KeyError:
            reason = data[LAVALINK_KEY_EXCEPTION_REASON_DEPRECATED]
            severity = ''
        else:
            reason = reason_object[LAVALINK_KEY_EXCEPTION_REASON_OBJECT_MESSAGE]
            severity = reason_object[LAVALINK_KEY_EXCEPTION_REASON_OBJECT_SEVERITY]
        
        track = player.get_current()
        
        self = object.__new__(cls)
        self.player = player
        self.track = track
        self.reason = reason
        self.severity = severity
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = []
        
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        repr_parts.append(' player = ')
        repr_parts.append(repr(self.player))
        
        repr_parts.append(', track = ')
        repr_parts.append(repr(self.track))
        
        severity = self.severity
        if severity:
            repr_parts.append(', severity = ')
            repr_parts.append(repr(severity))
        
        repr_parts.append(', reason = ')
        repr_parts.append(reprlib.repr(self.reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.player
        yield self.track
        yield self.reason
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.player != other.player:
            return False
        
        if self.reason != other.reason:
            return False
        
        if self.severity != other.severity:
            return False
        
        if self.track != other.track:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # player
        hash_value ^= hash(self.player)
        
        # reason
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        # severity
        severity = self.severity
        if severity:
            hash_value ^= hash(severity)
        
        # track
        hash_value ^= hash(self.track)
        
        return hash_value


class PlayerWebSocketClosedEvent(EventBase):
    """
    Represents an event when a player's web socket is disconnected from a guild.
    
    It usually is not a lavalink issue, but an issue with the stream.
    
    Attributes
    ----------
    by_remote : `bool`
        Whether the web socket was closed remotely.
    code : `int`
        Websocket close code.
    player : ``SolarPlayerBase``
        The player associated with the event.
    reason : `None`, `str`
        Websocket close reason.
    """
    __slots__ = ('by_remote', 'code', 'player', 'reason', )
    
    def __new__(cls, player, data):
        """
        Creates a new ``PlayerWebSocketClosedEvent`` event from the given parameters.
        
        Parameters
        ----------
        player : ``SolarPlayerBase``
            The player associated with the event.
        data : `dict` of (`str`, `object`) items
            Event payload.
        """
        close_code = data[LAVALINK_KEY_WEBSOCKET_CLOSE_CODE]
        close_reason = data[LAVALINK_KEY_WEBSOCKET_CLOSE_REASON]
        close_by_remote = data[LAVALINK_KEY_WEBSOCKET_CLOSE_BY_REMOTE]
        
        self = object.__new__(cls)
        self.player = player
        self.code = close_code
        self.reason = close_reason
        self.by_remote = close_by_remote
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = []
        
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        repr_parts.append(' player = ')
        repr_parts.append(repr(self.player))
        
        repr_parts.append(', code = ')
        repr_parts.append(repr(self.code))
        
        reason = self.reason
        if (reason is not None):
            repr_parts.append(', reason = ')
            repr_parts.append(reprlib.repr(reason))
        
        by_remote = self.by_remote
        if by_remote:
            repr_parts.append(', by_remote = ')
            repr_parts.append(repr(by_remote))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 4
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.player
        yield self.code
        yield self.reason
        yield self.by_remote
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.by_remote != other.by_remote:
            return False
        
        if self.code != other.code:
            return False
        
        if self.player != other.player:
            return False
        
        if self.reason != other.reason:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # by_remote
        hash_value ^= self.by_remote
        
        # code
        hash_value ^= (self.code << 8)
        
        # player
        hash_value ^= hash(self.player)
        
        # reason
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        return hash_value
    
    
    def should_reconnect(self):
        """
        Return whether the client should reconnect even against the closage.
        
        This method is based mostly on experimentations.
        
        Returns
        -------
        should_reconnect : `bool`
        """
        if not self.by_remote:
            if self.code == 1006:
                if self.reason == 'Abnormal closure':
                    return True
        
        return False



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, nullable_entity_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, force_string_putter_factory,
    nullable_entity_array_putter_factory, preinstanced_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_entity_array_validator_factory, preinstanced_validator_factory
)

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import ConnectionType, ConnectionVisibility


# parse_friend_sync

parse_friend_sync = bool_parser_factory('friend_sync', False)
put_friend_sync = bool_optional_putter_factory('friend_sync', False)
validate_friend_sync = bool_validator_factory('friend_sync', False)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('connection_id')

# integrations

parse_integrations = nullable_entity_array_parser_factory('integrations', NotImplemented, include = 'Integration')
put_integrations = nullable_entity_array_putter_factory(
    'integrations', NotImplemented, can_include_internals = True, include = 'Integration'
)
validate_integrations = nullable_entity_array_validator_factory('integrations', NotImplemented, include = 'Integration')

# metadata_visibility

parse_metadata_visibility = preinstanced_parser_factory(
    'metadata_visibility', ConnectionVisibility, ConnectionVisibility.user_only
)
put_metadata_visibility = preinstanced_optional_putter_factory(
    'metadata_visibility', ConnectionVisibility.user_only
)
validate_metadata_visibility = preinstanced_validator_factory('metadata_visibility', ConnectionVisibility)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# revoked

parse_revoked = bool_parser_factory('revoked', False)
put_revoked = bool_optional_putter_factory('revoked', False)
validate_revoked = bool_validator_factory('revoked', False)

# show_activity

parse_show_activity = bool_parser_factory('show_activity', False)
put_show_activity = bool_optional_putter_factory('show_activity', False)
validate_show_activity = bool_validator_factory('show_activity', False)

# two_way_link

parse_two_way_link = bool_parser_factory('two_way_link', False)
put_two_way_link = bool_optional_putter_factory('two_way_link', False)
validate_two_way_link = bool_validator_factory('two_way_link', False)

# type

parse_type = preinstanced_parser_factory('type', ConnectionType, ConnectionType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ConnectionType)

# verified

parse_verified = bool_parser_factory('verified', False)
put_verified = bool_optional_putter_factory('verified', False)
validate_verified = bool_validator_factory('verified', False)

# visibility

parse_visibility = preinstanced_parser_factory('visibility', ConnectionVisibility, ConnectionVisibility.user_only)
put_visibility = preinstanced_optional_putter_factory('visibility', ConnectionVisibility.user_only)
validate_visibility = preinstanced_validator_factory('visibility', ConnectionVisibility)



__all__ = ('ComponentMetadataSelectBase', )

import reprlib

from scarletio import copy_docs

from ..shared_fields import parse_custom_id, put_custom_id_into, validate_custom_id
from ..shared_helpers import create_auto_custom_id

from .base import ComponentMetadataBase
from .constants import MAX_VALUES_DEFAULT, MIN_VALUES_DEFAULT
from .fields import (
    parse_enabled, parse_max_values, parse_min_values, parse_placeholder, put_enabled, put_max_values,
    put_min_values, put_placeholder, validate_enabled, validate_max_values, validate_min_values,
    validate_placeholder
)


class ComponentMetadataSelectBase(ComponentMetadataBase):
    """
    Base select component metadata.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    """
    __slots__ = ('custom_id', 'enabled', 'max_values', 'min_values', 'placeholder')
    
    
    def __new__(cls, *, custom_id = ..., enabled = ..., max_values = ..., min_values = ..., placeholder = ...):
        """
        Creates a new base select component metadata with the given parameters.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # enabled
        if enabled is ...:
            enabled = True
        else:
            enabled = validate_enabled(enabled)
        
        # max_values
        if max_values is ...:
            max_values = MAX_VALUES_DEFAULT
        else:
            max_values = validate_max_values(max_values)
        
        # min_values
        if min_values is ...:
            min_values = MIN_VALUES_DEFAULT
        else:
            min_values = validate_min_values(min_values)
        
        # placeholder
        if placeholder is ...:
            placeholder = None
        else:
            placeholder = validate_placeholder(placeholder)
        
        # Extra checks
        
        if custom_id is None:
            custom_id = create_auto_custom_id()
        
        # Construct
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.enabled = enabled
        self.max_values = max_values
        self.min_values = min_values
        self.placeholder = placeholder
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            custom_id = keyword_parameters.pop('custom_id', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # System fields : custom_id
        
        # custom_id
        repr_parts.append(' custom_id = ')
        repr_parts.append(reprlib.repr(self.custom_id))
        
        # Type specific fields
        self._add_type_specific_repr_fields_into(repr_parts)
        
        # Text fields : placeholder
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            repr_parts.append(', placeholder = ')
            repr_parts.append(repr(placeholder))
        
        # Optional descriptive fields: min_values & max_values & enabled
        
        # min_values
        min_values = self.min_values
        if min_values != MIN_VALUES_DEFAULT:
            repr_parts.append(', min_values = ')
            repr_parts.append(repr(min_values))
        
        # max_values
        max_values = self.max_values
        if max_values != MAX_VALUES_DEFAULT:
            repr_parts.append(', max_values = ')
            repr_parts.append(repr(max_values))
        
        # enabled
        enabled = self.enabled
        if (not enabled):
            repr_parts.append(', enabled = ')
            repr_parts.append(repr(enabled))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _add_type_specific_repr_fields_into(self, repr_parts):
        """
        Adds type specific representation fields into the given list.
        
        Parameters
        ----------
        repr_parts : `list` of `str`
            Representation builder string to extend.
        """
        
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # enabled
        if self.enabled:
            hash_value ^= 1 << 8
        
        # max_values
        max_values = self.max_values
        if (max_values != 1):
            hash_value ^= (max_values << 18)
        
        # min_values
        min_values = self.min_values
        if (min_values != 1):
            min_values ^= (min_values << 22)
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            hash_value ^= hash(placeholder)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # max_values
        if self.max_values != other.max_values:
            return False
        
        # min_values
        if self.min_values != other.min_values:
            return False
        
        # placeholder
        if self.placeholder != other.placeholder:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.custom_id = parse_custom_id(data)
        self.enabled = parse_enabled(data)
        self.max_values = parse_max_values(data)
        self.min_values = parse_min_values(data)
        self.placeholder = parse_placeholder(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        
        put_custom_id_into(self.custom_id, data, defaults)
        put_enabled(self.enabled, data, defaults)
        put_max_values(self.max_values, data, defaults)
        put_min_values(self.min_values, data, defaults)
        put_placeholder(self.placeholder, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # custom_id
        new.custom_id = self.custom_id
        
        # enabled
        new.enabled = self.enabled
        
        # placeholder
        new.placeholder = self.placeholder
        
        # max_values
        new.max_values = self.max_values
        
        # min_values
        new.min_values = self.min_values
        
        return new
    
    
    def copy_with(
        self,
        *,
        channel_types = ...,
        custom_id = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        placeholder = ...,
    ):
        """
        Copies the base select component metadata with the given fields.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # enabled
        if enabled is ...:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # max_values
        if max_values is ...:
            max_values = self.max_values
        else:
            max_values = validate_max_values(max_values)
        
        # min_values
        if min_values is ...:
            min_values = self.min_values
        else:
            min_values = validate_min_values(min_values)
        
        # placeholder
        if placeholder is ...:
            placeholder = self.placeholder
        else:
            placeholder = validate_placeholder(placeholder)
        
        # Extra checks
        
        if custom_id is None:
            custom_id = create_auto_custom_id()
        
        # Construct
        new = object.__new__(type(self))
        new.custom_id = custom_id
        new.enabled = enabled
        new.max_values = max_values
        new.min_values = min_values
        new.placeholder = placeholder
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            custom_id = keyword_parameters.pop('custom_id', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )



__all__ = ('AudioSettings',)

from ctypes import c_char as CCharType

from scarletio import RichAttributeErrorBaseType

from .constants import CHANNELS_DEFAULT, FRAME_LENGTH_DEFAULT, SAMPLING_RATE_DEFAULT
from .fields import validate_channels, validate_frame_length, validate_sampling_rate


class AudioSettings(RichAttributeErrorBaseType):
    """
    Attributes
    ----------
    buffer_type : `type<CCharArrayType>`
        C char array type for creating buffer.
    
    channels : `int`
        The number of channels. (1 if mono, 2 if stereo.)
    
    frame_length : `int`
        The length of a frame in milliseconds.
    
    frame_size : `int`
        The size of a frame in bytes.
    
    sampling_rate : `int`
        The number of samples per second that are taken of a waveform to create a digital signal.
        The higher the sample rate, the more snapshots are captured in the audio signal.
    
    sample_size : `int`
        The number of bits used to describe each sample.
        A sample is a value represented in `i16`. Each channel has their own samples.
    
    samples_per_frame : `int`
        The amount of samples that are in each frame.
    """
    __slots__ = (
        'buffer_type', 'channels', 'frame_length', 'frame_size', 'samples_per_frame', 'sampling_rate', 'sample_size'
    )
    
    def __new__(
        cls,
        *,
        channels = ...,
        frame_length = ...,
        sampling_rate = ...,
    ):
        """
        Creates a new audio setting.
        
        Parameters
        ----------
        channels : `int`, Optional (Keyword only)
            The number of channels.
        frame_length : `int`, Optional (Keyword only)
            The length of a frame in milliseconds.
        sampling_rate : `int`, Optional (Keyword only)
            The number of samples per second that are taken of a waveform to create a digital signal.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channels
        if channels is ...:
            channels = CHANNELS_DEFAULT
        else:
            channels = validate_channels(channels)
        
        # frame_length
        if frame_length is ...:
            frame_length = FRAME_LENGTH_DEFAULT
        else:
            frame_length = validate_frame_length(frame_length)
        
        # sampling_rate
        if sampling_rate is ...:
            sampling_rate = SAMPLING_RATE_DEFAULT
        else:
            sampling_rate = validate_sampling_rate(sampling_rate)
        
        # Do math
        sample_size = 2 * channels # sizeof(i16) * channels
        samples_per_frame = int(sampling_rate / 1000 * frame_length)
        frame_size = samples_per_frame * sample_size
        buffer_type = CCharType * frame_size
        
        # Construct
        self = object.__new__(cls)
        self.buffer_type = buffer_type
        self.channels = channels
        self.frame_length = frame_length
        self.frame_size = frame_size
        self.samples_per_frame = samples_per_frame
        self.sampling_rate = sampling_rate
        self.sample_size = sample_size
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two audio settings are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.channels != other.channels:
            return False
        
        if self.frame_length != other.frame_length:
            return False
        
        if self.sampling_rate != other.sampling_rate:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the audio settings representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' channels = ')
        repr_parts.append(repr(self.channels))
        
        repr_parts.append(', frame_length = ')
        repr_parts.append(repr(self.frame_length))
        
        repr_parts.append(', sampling_rate = ')
        repr_parts.append(repr(self.sampling_rate))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the audio settings.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.buffer_type = self.buffer_type
        new.channels = self.channels
        new.frame_length = self.frame_length
        new.frame_size = self.frame_size
        new.samples_per_frame = self.samples_per_frame
        new.sampling_rate = self.sampling_rate
        new.sample_size = self.sample_size
        return new
    
    
    def copy_with(
        self,
        *,
        channels = ...,
        frame_length = ...,
        sampling_rate = ...,
    ):
        """
        Copies the audio settings with the given fields.
        
        Parameters
        ----------
        channels : `int`, Optional (Keyword only)
            The number of channels.
        frame_length : `int`, Optional (Keyword only)
            The length of a frame in milliseconds.
        sampling_rate : `int`, Optional (Keyword only)
            The number of samples per second that are taken of a waveform to create a digital signal.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channels
        if channels is ...:
            channels = self.channels
        else:
            channels = validate_channels(channels)
        
        # frame_length
        if frame_length is ...:
            frame_length = self.frame_length
        else:
            frame_length = validate_frame_length(frame_length)
        
        # sampling_rate
        if sampling_rate is ...:
            sampling_rate = self.sampling_rate
        else:
            sampling_rate = validate_sampling_rate(sampling_rate)
        
        # Do math
        sample_size = 2 * channels # sizeof(i16) * channels
        samples_per_frame = int(sampling_rate / 1000 * frame_length)
        frame_size = samples_per_frame * sample_size
        buffer_type = CCharType * frame_size
        
        # Construct
        new = object.__new__(type(self))
        new.buffer_type = buffer_type
        new.channels = channels
        new.frame_length = frame_length
        new.frame_size = frame_size
        new.samples_per_frame = samples_per_frame
        new.sampling_rate = sampling_rate
        new.sample_size = sample_size
        return new



__all__ = ()

from ...field_parsers import nullable_object_array_parser_factory, nullable_string_parser_factory
from ...field_putters import (
    force_bool_putter_factory, nullable_object_array_optional_putter_factory, nullable_string_putter_factory
)
from ...field_validators import (
    bool_validator_factory, nullable_object_array_validator_factory, nullable_string_validator_factory
)

from ..welcome_screen_channel import WelcomeScreenChannel

from .constants import DESCRIPTION_LENGTH_MAX

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# enabled

put_enabled = force_bool_putter_factory('enabled')
validate_enabled = bool_validator_factory('enabled', True)

# welcome_channels

parse_welcome_channels = nullable_object_array_parser_factory('welcome_channels', WelcomeScreenChannel)
put_welcome_channels = nullable_object_array_optional_putter_factory('welcome_channels')
validate_welcome_channels = nullable_object_array_validator_factory('welcome_channels', WelcomeScreenChannel)



__all__ = ()


# expire grace period
EXPIRE_GRACE_PERIOD_DEFAULT = 0

EXPIRE_GRACE_PERIOD_OPTIONS = frozenset((
    EXPIRE_GRACE_PERIOD_DEFAULT,
    1, 3, 7, 14, 30
))

# subscriber count
SUBSCRIBER_COUNT_DEFAULT = 0



__all__ = ()

from .base import DiscordGatewayBase


class DiscordGatewayClientBase(DiscordGatewayBase):
    """
    Base gateway for clients.
    """
    __slots__ = ()
    
    async def change_voice_state(self, guild_id, channel_id, *, self_deaf = False, self_mute = False):
        """
        Sends a `VOICE_STATE` packet to Discord.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild_id : `int`
            The voice client's guild's id.
        channel_id : `int`
            The voice client's channel's id.
        self_deaf : `bool` = `False`, Optional (Keyword only)
            Whether the voice client is deafen.
        self_mute : `bool` = `False`, Optional (Keyword only)
            Whether the voice client is muted.
        """
        pass
    
    
    def get_gateway(self, guild_id):
        """
        Returns the gateway for the given `guild_id`.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's identifier to get the gateway for. Can be `0`.
        
        Returns
        -------
        gateway : ``DiscordGatewayClientBase``
        """
        return self



__all__ = ('AuditLogRole',)

from ...bases import DiscordEntity
from ...role import Role

from .fields import parse_id, parse_name, put_id, put_name, validate_id, validate_name


class AuditLogRole(DiscordEntity):
    """
    Represents a role object received with audit logs.
    
    Parameters
    ----------
    id : `int`
        The role's identifier.
    name : `str`
        The role's name.
    """
    __slots__ = ('name',)
    
    def __new__(cls, *, role_id = ..., name = ...):
        """
        Creates a new new audit log role with the given parameters.
        
        Parameters
        ----------
        role_id : `int`, Optional (Keyword only)
            The represented role's identifier.
        name : `str`, Optional (Keyword only)
            The name of the role.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # role_id
        if role_id is ...:
            role_id = 0
        else:
            role_id = validate_id(role_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Construct
        self = object.__new__(cls)
        self.id = role_id
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new audit log role.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Audit log role data.
        """
        role_id = parse_id(data)
        name = parse_name(data)
        
        self = object.__new__(cls)
        self.id = role_id
        self.name = name
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the audit log role field.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_id(self.id, data, defaults)
        put_name(self.name, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the audit log role's representation."""
        repr_parts = ['<',self.__class__.__name__]
        
        repr_parts.append(' id = ')
        repr_parts.append(repr(self.id))
        
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two audit log role's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two audit log role's are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        if self.id != other.id:
            return False
        
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """returns the audit log role's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= hash(self.id)
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the audit log role returning a new one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = self.id
        new.name = self.name
        return new
    
    
    def copy_with(self, *, role_id = ..., name = ...):
        """
        Copies the audit log role with the given fields.
        
        Parameters
        ----------
        role_id : `int`, Optional (Keyword only)
            The represented role's identifier.
        name : `str`, Optional (Keyword only)
            The name of the role.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # role_id
        if role_id is ...:
            role_id = self.id
        else:
            role_id = validate_id(role_id)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Construct
        new = object.__new__(type(self))
        new.id = role_id
        new.name = name
        return new
    
    
    @property
    def entity(self):
        """
        Resolves the audit log role's entity.
        
        If the entity is not cached creates a new.
        
        Returns
        -------
        role : ``Role``
        """
        return Role.precreate(self.id, name = self.name)



__all__ = ('SolarAuthenticationError', 'SolarException', )

class SolarException(Exception):
    """
    Base class for solar client exceptions.
    """

class SolarAuthenticationError(SolarException):
    """
    Raises when the client could not authorize itself a lavalink node.
    
    Attributes
    ----------
    node : ``SolarNode``
        The node we tried to connect to.
    response : ``ClientResponse``
        The http response received from the node.
    """
    def __init__(self, node, response):
        self.node = node
        self.response = response
        Exception.__init__(self, node, response)
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'<{self.__class__.__name__} node={self.node!r}, response={self.response!r}>'



__all__ = ()

from ...field_parsers import entity_id_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, force_string_putter_factory,
    nullable_string_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_string_validator_factory
)

from ..sticker import Sticker

# banner_id

parse_banner_id = entity_id_parser_factory('banner_asset_id')
put_banner_id = entity_id_optional_putter_factory('banner_asset_id')
validate_banner_id = entity_id_validator_factory('banner_id')

# cover_sticker_id

parse_cover_sticker_id = entity_id_parser_factory('cover_sticker_id')
put_cover_sticker_id = entity_id_optional_putter_factory('cover_sticker_id')
validate_cover_sticker_id = entity_id_validator_factory('cover_sticker_id', Sticker)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, 1024)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, 1024)

# sku_id

parse_sku_id = entity_id_parser_factory('sku_id')
put_sku_id = entity_id_optional_putter_factory('sku_id')
validate_sku_id = entity_id_validator_factory('sku_id', NotImplemented, include = 'SKU')

# stickers

def parse_stickers(data):
    """
    Parses the stickers out from the given sticker-pack data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Sticker-pack data.
    
    Returns
    -------
    stickers : `None``, `frozenset` of ``Sticker``
    """
    sticker_datas = data.get('stickers', None)
    if (sticker_datas is not None) and sticker_datas:
        return frozenset(Sticker.from_data(sticker_data) for sticker_data in sticker_datas)


def put_stickers(stickers, data, defaults):
    """
    Puts the stickers into the given json serializable dictionary.
    
    Parameters
    ----------
    stickers : `None`, `frozenset` of `Sticker`
        Stickers.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if stickers is None:
        sticker_datas = []
    else:
        sticker_datas = [sticker.to_data(defaults = defaults, include_internals = True) for sticker in stickers]
    data['stickers'] = sticker_datas
    return data


def validate_stickers(stickers):
    """
    Validates the given stickers.
    
    Parameters
    ----------
    stickers : `None`, `iterable` of ``Sticker``
        The stickers to validate.
    
    Returns
    -------
    stickers : `None`, `frozenset` of ``Sticker``
        The validated stickers.
    
    Raises
    ------
    TypeError
        - If `sticker` is not `None`, `iterable` of ``Sticker``.
        - If `sticker` contains a non-``Sticker`` element. 
    """
    if stickers is None:
        return None
    
    if getattr(type(stickers), '__iter__', None) is None:
        raise TypeError(
            f'`stickers` can be `None` ot `iterable` of `{Sticker.__name__}`, '
            f'got {stickers.__class__.__name__}; {stickers!r}.'
        )
    
    validated_stickers = None
    
    for sticker in stickers:
        if not isinstance(sticker, Sticker):
            raise TypeError(
                f'`sticker` can contain `{Sticker.__name__}` elements, got {sticker.__class__.__name__}; {sticker!r}; '
                f'stickers = {stickers!r}.'
            )
        
        if (validated_stickers is None):
            validated_stickers = set()
        
        validated_stickers.add(sticker)
    
    if (validated_stickers is not None):
        return frozenset(validated_stickers)



__all__ = ()

from ...activity import Activity
from ...color import Color
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, nullable_entity_parser_factory,
    nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_putter_factory, force_bool_putter_factory,
    force_string_putter_factory, nullable_entity_optional_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    nullable_entity_validator_factory, nullable_string_validator_factory, preinstanced_validator_factory
)
from ...localization import Locale
from ...localization.utils import LOCALE_DEFAULT

from ..avatar_decoration import AvatarDecoration
from ..user_clan import UserClan

from .constants import (
    DISCRIMINATOR_VALUE_MAX, DISCRIMINATOR_VALUE_MIN, DISPLAY_NAME_LENGTH_MAX, NAME_LENGTH_MAX, NAME_LENGTH_MIN,
    WEBHOOK_NAME_LENGTH_MAX, WEBHOOK_NAME_LENGTH_MIN
)
from .flags import UserFlag
from .preinstanced import PremiumType, Status

# activities

def parse_activities(data):
    """
    Parses the activities out from the given presence data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User presence data.
    
    Returns
    -------
    activities : `None`, `list` of ``Activity``
    """
    activity_datas = data.get('activities', None)
    if (activity_datas is None) or (not activity_datas):
        return None
    
    return [Activity.from_data(activity_data) for activity_data in activity_datas]


def put_activities(activities, data, defaults):
    """
    Puts the given activities into the given data.
    
    Parameters
    ----------
    activities : `None`, `list` of ``Activity``
        Activities.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if activities is None:
        activity_datas = []
    else:
        activity_datas = [
            activity.to_data(defaults = defaults, include_internals = True, user = True)
            for activity in activities
        ]
    
    data['activities'] = activity_datas
    return data


def validate_activities(activities):
    """
    Validates the given activities.
    
    Parameters
    ----------
    activities : `None`, `iterable` of ``Activity``
        The activities to validate.
    
    Returns
    -------
    activities : `None`, `list` of ``Activity``
    
    Raises
    ------
    TypeError
        - If `activities` type is incorrect.
    """
    if activities is None:
        return None
    
    if getattr(activities, '__iter__', None) is None:
        raise TypeError(
            f'`activities` can be `None`, `iterable` of `{Activity.__name__}`, got '
            f'{activities.__class__.__name__}; {activities!r}.'
        )
    
    activities_validated = None
    
    for entity in activities:
        if not isinstance(entity, Activity):
            raise TypeError(
                f'`activities` elements can be `{Activity.__name__}`, got '
                f'{entity.__class__.__name__}; {entity!r}; activities = {activities!r}'
            )
        
        if activities_validated is None:
            activities_validated = []
        
        activities_validated.append(entity)
    
    return activities_validated

# avatar_decoration

parse_avatar_decoration = nullable_entity_parser_factory('avatar_decoration_data', AvatarDecoration)
put_avatar_decoration = nullable_entity_optional_putter_factory('avatar_decoration_data', AvatarDecoration)
validate_avatar_decoration = nullable_entity_validator_factory('avatar_decoration', AvatarDecoration)

# banner_color

def parse_banner_color(data):
    """
    Gets banner color from the given user data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User data.
    
    Returns
    -------
    banner_color : `None`, ``Color``
    """
    banner_color = data.get('accent_color', None)
    if (banner_color is not None):
        banner_color = Color(banner_color)
    
    return banner_color


def put_banner_color(banner_color, data, defaults):
    """
    Puts the given banner color into the given data.
    
    Parameters
    ----------
    banner_color : `None`, ``Color``
        Banner color.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (banner_color is not None):
        if (banner_color is not None):
            banner_color = int(banner_color)
        data['accent_color'] = banner_color
    
    return data


def validate_banner_color(banner_color):
    """
    Validates the given `banner_color` value.
    
    Parameters
    ----------
    banner_color : `None`, ``Color``, `int`
        The banner color to validate.
    
    Returns
    -------
    banner_color : `None`, ``Color``
    
    Raises
    ------
    TypeError
        - If `banner_color` is not `None`, `Color`, `int`.
    ValueError
        - If `banner_color`'s value is out of the expected range.
    """
    if (banner_color is None):
        return None
    
    if isinstance(banner_color, Color):
        return banner_color
    
    if isinstance(banner_color, int):
        if banner_color < 0 or banner_color > 0xffffff:
            raise ValueError(
                f'`banner_color` can be between 0 and 0xffffff, got {banner_color!r}.'
            )
        
        return Color(banner_color)
    
    raise TypeError(
        f'`banner_color` can be `None`, `{Color.__name__}`, `int`, '
        f'got {type(banner_color).__name__}; {banner_color!r}.'
    )

# bot

parse_bot = bool_parser_factory('bot', False)
put_bot = force_bool_putter_factory('bot')
validate_bot = bool_validator_factory('bot', False)

# clan

parse_clan = nullable_entity_parser_factory('clan', UserClan)
put_clan = nullable_entity_optional_putter_factory('clan', UserClan)
validate_clan = nullable_entity_validator_factory('clan', UserClan)

# discriminator

parse_discriminator = flag_parser_factory('discriminator', int)

def put_discriminator(discriminator, data, defaults):
    """
    Puts the given discriminator value into the given data.
    
    Parameters
    ----------
    discriminator : `int`
        Discriminator value.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['discriminator'] = str(discriminator).rjust(4, '0')
    return data


def validate_discriminator(discriminator):
    """
    validates the given discriminator.
    
    Parameters
    ----------
    discriminator : `str`, `int`
        The discriminator of an user to convert.
    
    Returns
    -------
    discriminator : `int`
    
    Raises
    ------
    TypeError
        - If `discriminator` was not passed neither as `int`, `str`.
    ValueError
        - If `discriminator` was passed as `str` and it is not numerical or it's length is over `4`.
        - If the `discriminator`'s value is less than `0` or is over than `9999`.
    """
    if (type(discriminator) is int):
        pass
    
    elif isinstance(discriminator, int):
        discriminator = int(discriminator)
    
    elif isinstance(discriminator, str):
        if not discriminator.isdigit():
            raise ValueError(
                f'`discriminator` can contain only numerical characters (0 - 9), got {discriminator!r}.'
            )
        discriminator = int(discriminator)
    
    else:
        raise TypeError(
            f'`discriminator` can be `int`, `str`, got '
            f'{discriminator.__class__.__name__}; {discriminator!r}.'
        )
    
    if (discriminator < DISCRIMINATOR_VALUE_MIN) or (discriminator > DISCRIMINATOR_VALUE_MAX):
        raise ValueError(
            f'`discriminator` must be >= {DISCRIMINATOR_VALUE_MIN!r} and <= {DISCRIMINATOR_VALUE_MAX!r}, '
            f'got {discriminator!r}.'
        )
    
    return discriminator


# display_name

parse_display_name = nullable_string_parser_factory('global_name')
put_display_name = url_optional_putter_factory('global_name')
validate_display_name = nullable_string_validator_factory('global_name', 0, DISPLAY_NAME_LENGTH_MAX)

# email

parse_email = nullable_string_parser_factory('email')
put_email = url_optional_putter_factory('email')
validate_email = nullable_string_validator_factory('email', 0, 1024)

# email_verified

parse_email_verified = bool_parser_factory('verified', False)
put_email_verified = bool_optional_putter_factory('verified', False)
validate_email_verified = bool_validator_factory('email_verified', False)

# flags

def parse_flags(data):
    """
    Parses the user's flags out from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User data.
    
    Returns
    -------
    flags : ``UserFlag``
    """
    try:
        flags = data['flags']
    except KeyError:
        flags = data.get('public_flags', 0)
    
    return UserFlag(flags)

put_flags = flag_putter_factory('public_flags')
validate_flags = flag_validator_factory('public_flags', UserFlag)

# flags | oauth2_flags

put_oauth2_flags_into = flag_putter_factory('flags')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('user_id')

# locale

parse_locale = preinstanced_parser_factory('locale', Locale, LOCALE_DEFAULT)
put_locale = preinstanced_putter_factory('locale')
validate_locale = preinstanced_validator_factory('locale', Locale)

# mfa_enabled

parse_mfa_enabled = bool_parser_factory('mfa_enabled', False)
put_mfa_enabled = bool_optional_putter_factory('mfa_enabled', False)
validate_mfa_enabled = bool_validator_factory('mfa_enabled', False)

# name

def parse_name(data):
    """
    Parses out the given user's name from the data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User data.
    
    Returns
    -------
    name : `str`
    """
    try:
        name = data['username']
    except KeyError:
        # Webhook?
        name = data.get('name', None)
    
    if (name is None):
        name = ''
        
    return name


put_name = force_string_putter_factory('username')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# name | webhook_name

put_webhook_name = force_string_putter_factory('name')
validate_webhook_name = force_string_validator_factory('name', WEBHOOK_NAME_LENGTH_MIN, WEBHOOK_NAME_LENGTH_MAX)

# premium_type

parse_premium_type = preinstanced_parser_factory('premium_type', PremiumType, PremiumType.none)
put_premium_type = preinstanced_putter_factory('premium_type')
validate_premium_type = preinstanced_validator_factory('premium_type', PremiumType)

# status

parse_status = preinstanced_parser_factory('status', Status, Status.offline)
put_status = preinstanced_putter_factory('status')
validate_status = preinstanced_validator_factory('status', Status)

# statuses

def parse_statuses(data):
    """
    Parses out a user statuses from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User presence data.
    
    Returns
    -------
    statuses : `None`, `dict` of (`str`, `str`) items
    """
    statuses = data.get('client_status', None)
    if (statuses is not None) and (not statuses):
        statuses = None
    
    return statuses


def put_statuses(statuses, data, defaults):
    """
    Puts the given statuses value into the given data.
    
    Parameters
    ----------
    statuses : `None`, `dict` of (`str`, `str`) items
        user statuses by platform.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (statuses is None):
        statuses = {}
    
    data['client_status'] = statuses
    return data


def validate_statuses(statuses):
    """
    validates the given `statuses` value.
    
    Parameters
    ----------
    statuses : `None`, `dict` of (`str`, `str`) items
        Statuses to validate.
    
    Returns
    -------
    statuses : `None`, `dict` of (`str`, `str`) items
    
    Raises
    ------
    TypeError
        - If `statuses`'s type is incorrect.
    """
    if statuses is None:
        return None
    
    if not isinstance(statuses, dict):
        raise TypeError(
            f'`statuses` can be `None` or `dict` of (`str`, `str`) items, '
            f'got {type(statuses).__name__}; {statuses!r}.'
        )
    
    statuses_validated = None
    
    for key, value in statuses.items():
        if not isinstance(key, str):
            raise TypeError(
                f'`statuses` keys can be `str`, got {key.__class__.__name__}; {key!r}; statuses = {statuses!r}.'
            )
        if not isinstance(value, str):
            raise TypeError(
                f'`statuses` values can be `str`, got {value.__class__.__name__}; {value!r}; statuses = {statuses!r}.'
            )
        
        if statuses_validated is None:
            statuses_validated = {}
            
        statuses_validated[key] = value
    
    return statuses_validated



__all__ = ()

from functools import partial as partial_func

from .application import APPLICATION_COVER, APPLICATION_ICON
from .fields import (
    put_custom_install_url, put_description, put_flags, put_install_parameters,
    put_interaction_endpoint_url, put_role_connection_verification_url, put_tags,
    validate_custom_install_url, validate_description, validate_flags, validate_install_parameters,
    validate_interaction_endpoint_url, validate_role_connection_verification_url, validate_tags
)


APPLICATION_FIELD_CONVERTERS = {
    'cover': (
        partial_func(APPLICATION_COVER.validate_icon, allow_data = True),
        partial_func(APPLICATION_COVER.put_into, as_data = True),
    ),
    'custom_install_url': (validate_custom_install_url, put_custom_install_url),
    'description': (validate_description, put_description),
    'flags': (validate_flags, put_flags),
    'icon': (
        partial_func(APPLICATION_ICON.validate_icon, allow_data = True),
        partial_func(APPLICATION_ICON.put_into, as_data = True),
    ),
    'install_parameters': (validate_install_parameters, put_install_parameters),
    'interaction_endpoint_url': (validate_interaction_endpoint_url, put_interaction_endpoint_url),
    'role_connection_verification_url': (
        validate_role_connection_verification_url, put_role_connection_verification_url,
    ),
    'tags': (validate_tags, put_tags),
}



__all__ = ()

import reprlib

from scarletio import Compound, IgnoreCaseMultiValueDictionary
from scarletio.web_common.headers import AUTHORIZATION

from ....env import API_VERSION

from ...application import Application
from ...http import DiscordApiClient, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...localization.helpers import serializable_localized_dictionary_builder
from ...oauth2 import Achievement, Oauth2Access, Oauth2User
from ...utils import get_image_media_type, image_to_base64
from ..request_helpers import get_achievement_and_id, get_achievement_id, get_user_id


def _assert__achievement_create__name(name):
    """
    Asserts the `name` parameter of ``Client.achievement_create`` methods.
    
    Parameters
    ----------
    name : `str`
        The achievement's name.
    
    Raises
    ------
    AssertionError
        - If `name` was not given as `str`.
    """
    if not isinstance(name, str):
        raise AssertionError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    return True


def _assert__achievement_create__description(description):
    """
    Asserts the `description` parameter of ``Client.achievement_create`` methods.
    
    Parameters
    ----------
    description : `str`
        The achievement's description.
    
    Raises
    ------
    AssertionError
        - If `description` was not given as `str`.
    """
    if not isinstance(description, str):
        raise AssertionError(
            f'`description` can be `str`, got {description.__class__.__name__}; {description!r}.'
        )
    
    return True


def _assert__achievement_create__secret(secret):
    """
    Asserts the `secret` parameter of ``Client.achievement_create`` methods.
    
    Parameters
    ----------
    secret : `bool`
        The achievement's secret.
    
    Raises
    ------
    AssertionError
        - If `secret` was not given as `bool`.
    """
    if not isinstance(secret, bool):
        raise AssertionError(
            f'`secret` can be `bool`, got {secret.__class__.__name__}; {secret!r}.'
        )
    
    return True


def _assert__achievement_create__secure(secure):
    """
    Asserts the `secure` parameter of ``Client.achievement_create`` methods.
    
    Parameters
    ----------
    secure : `bool`
        The achievement's secure.
    
    Raises
    ------
    AssertionError
        - If `secure` was not given as `bool`.
    """
    if not isinstance(secure, bool):
        raise AssertionError(
            f'`secure` can be `bool`, got {secure.__class__.__name__}; {secure!r}.'
        )
    
    return True


def _assert__achievement_edit__name(name):
    """
    Asserts the `name` parameter of ``Client.achievement_edit`` methods.
    
    Parameters
    ----------
    name : `str`
        The achievement's name.
    
    Raises
    ------
    AssertionError
        - If `name` was not given as `str`.
    """
    if (name is not ...) and (not isinstance(name, str)):
        raise AssertionError(
            f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
        )
    
    return True


def _assert__achievement_edit__description(description):
    """
    Asserts the `description` parameter of ``Client.achievement_edit`` methods.
    
    Parameters
    ----------
    description : `str`
        The achievement's description.
    
    Raises
    ------
    AssertionError
        - If `description` was not given as `str`.
    """
    if (description is not ...) and (not isinstance(description, str)):
        raise AssertionError(
            f'`description` can be `str`, got {description.__class__.__name__}; {description!r}.'
        )
    
    return True


def _assert__achievement_edit__secret(secret):
    """
    Asserts the `secret` parameter of ``Client.achievement_edit`` methods.
    
    Parameters
    ----------
    secret : `bool`
        The achievement's secret.
    
    Raises
    ------
    AssertionError
        - If `secret` was not given as `bool`.
    """
    if (secret is not ...) and (not isinstance(secret, bool)):
        raise AssertionError(
            f'`secret` can be `bool`, got {secret.__class__.__name__}; {secret!r}.'
        )
    
    return True


def _assert__achievement_edit__secure(secure):
    """
    Asserts the `secure` parameter of ``Client.achievement_edit`` methods.
    
    Parameters
    ----------
    secure : `bool`
        The achievement's secure.
    
    Raises
    ------
    AssertionError
        - If `secure` was not given as `bool`.
    """
    if (secure is not ...) and (not isinstance(secure, bool)):
        raise AssertionError(
            f'`secure` can be `bool`, got {secure.__class__.__name__}; {secure!r}.'
        )
    
    return True


def _assert__user_achievement_update__percent_complete(percent_complete):
    """
    Asserts the `percent_complete` parameter of ``Client.user_achievement_update`` methods.
    
    Parameters
    ----------
    percent_complete : `int`
        The achievement's percent_complete.
    
    Raises
    ------
    AssertionError
        - If `percent_complete` was not given as `int`.
    """
    if not isinstance(percent_complete, int):
        raise AssertionError(
            f'`percent_complete` can be `int`, got {percent_complete.__class__.__name__}; {percent_complete!r}.'
        )
    
    return True


class ClientCompoundAchievementEndpoints(Compound):
    
    application : Application
    api : DiscordApiClient
    
    
    async def achievement_get_all(self):
        """
        Requests all the achievements of the client's application and returns them.
        
        This method is a coroutine.
        
        Returns
        -------
        achievements : `list` of ``Achievement`` objects
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.api.achievement_get_all(self.application.id)
        return [Achievement(achievement_data) for achievement_data in data]
    
    
    async def achievement_get(self, achievement):
        """
        Requests one of the client's achievements by it's id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        achievement : ``Achievement``, `int`
            The achievement or it's identifier.
        
        Returns
        -------
        achievement : ``Achievement``
        
        Raises
        ------
        TypeError
            If `achievement` is not given as ``Achievement``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        achievement, achievement_id = get_achievement_and_id(achievement)
        data = await self.api.achievement_get(self.application.id, achievement_id)
        
        if achievement is None:
            achievement = Achievement(data)
        else:
            achievement._update_attributes(data)
        
        return achievement
    
    
    async def achievement_create(
        self,
        name,
        description,
        icon,
        *,
        description_localizations = None,
        name_localizations = None,
        secret = False,
        secure = False,
    ):
        """
        Creates an achievement for the client's application and returns it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`
            The achievement's name.
        description : `str`
            The achievement's description.
        icon : `bytes-like`
            The achievement's icon. Can have `'jpg'`, `'png'`, `'webp'`, `'gif'` format.
        description_localizations : `None`, `dict` of (`str`, `str`) items = `None`, Optional (Keyword only)
            Localized descriptions of the achievement.
        name_localizations : `None`, `dict` of (`str`, `object`) items = `None`, Optional (Keyword only)
            Localized names of the achievement.
        secret : `bool` = `False`, Optional (Keyword only)
            Secret achievements will *not* be shown to the user until they've unlocked them.
        secure : `bool` = `False`, Optional (Keyword only)
            Secure achievements can only be set via HTTP calls from your server, not by a game client using the SDK.
        
        Returns
        -------
        achievement : ``Achievement``
            The created achievement entity.
        
        Raises
        ------
        TypeError
            - If `icon` was not passed as `bytes-like`.
        ValueError
            - If `icon`'s format is not any of the expected ones.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__achievement_create__name(name)
        assert _assert__achievement_create__description(description)
        assert _assert__achievement_create__secret(secret)
        assert _assert__achievement_create__secure(secure)
        
        if not isinstance(icon, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`icon` can be `bytes-like`, got {icon.__class__.__name__}; {reprlib.repr(icon)}.'
            )
        
        media_type = get_image_media_type(icon)
        if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
            raise ValueError(
                f'Invalid `icon` type for achievement: {media_type}; got {reprlib.repr(icon)}.'
            )
        
        icon_data = image_to_base64(icon)
        
        
        description_localizations = serializable_localized_dictionary_builder(
            description_localizations, 'description_localizations'
        )
        if description_localizations is None:
            description_localizations = {}
        
        name_localizations = serializable_localized_dictionary_builder(name_localizations, 'name_localizations')
        if name_localizations is None:
            name_localizations = {}
        
        data = {
            'secret': secret,
            'secure': secure,
            'icon': icon_data,
        }
        
        
        if API_VERSION >= 10:
            data['name'] = name
            data['description'] = description
            data['name_localizations'] = name_localizations
            data['description_localizations'] = description_localizations
        
        else:
            name_localizations['default'] = name
            data['name'] = name_localizations
            
            description_localizations['default'] = description
            data['description'] = description_localizations
        
        data = await self.api.achievement_create(self.application.id, data)
        
        return Achievement(data)
    
    
    async def achievement_edit(
        self,
        achievement,
        *,
        description = ...,
        description_localizations = ...,
        icon = ...,
        name = ...,
        name_localizations = ...,
        secret = ...,
        secure = ...,
    ):
        """
        Edits the passed achievement with the specified parameters. All parameter is optional.
        
        This method is a coroutine.
        
        Parameters
        ----------
        achievement : ``Achievement``, `int`
            The achievement, what will be edited.
        description : `str`, Optional (Keyword only)
            The achievement's new description.
        description_localizations : `dict` of (`str`, `str`) items
            Localized descriptions of the achievement.
        icon : `bytes-like`, Optional (Keyword only)
            The achievement's new icon.
        name : `str`, Optional (Keyword only)
            The new name of the achievement.
        name_localizations : `dict` of (`str`, `str`) items
            Localized names of the achievement.
        secret : `bool`, Optional (Keyword only)
            The achievement's new secret value.
        secure : `bool`, Optional (Keyword only)
            The achievement's new secure value.
        
        Returns
        -------
        achievement : ``Achievement``
            After a successful edit, the passed achievement is updated and returned.
        
        Raises
        ------
        TypeError
            - If ``icon`` was not passed as `bytes-like`.
            - If `achievement` was not given neither as ``Achievement``, neither as `int`.
        ValueError
            - If `icon`'s format is not any of the expected ones.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        achievement, achievement_id = get_achievement_and_id(achievement)
        
        assert _assert__achievement_edit__name(name)
        assert _assert__achievement_edit__description(description)
        assert _assert__achievement_edit__secret(secret)
        assert _assert__achievement_edit__secure(secure)
        
        data = {}
        

        if (icon is not ...):
            if not isinstance(icon, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`icon` can be `bytes-like`, got {icon.__class__.__name__}; {reprlib.repr(icon)}.'
                )
            
            media_type = get_image_media_type(icon)
            if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                raise ValueError(
                    f'Invalid `icon` type for achievement: {media_type}; got {reprlib.repr(icon)!r}.'
                )
            
            data['icon'] = image_to_base64(icon)
        
        
        if (secret is not ...):
            data['secret'] = secret
        
        if (secure is not ...):
            data['secure'] = secure
        
       
        if API_VERSION >= 10:
            if (description is not ...):
                data['description'] = description
            
            
                description_localizations = serializable_localized_dictionary_builder(
                    description_localizations,
                    'description_localizations',
                )
                if description_localizations is None:
                    description_localizations = {}
                
                data['description_localizations'] = description_localizations
            
            
            if (name is not ...):
                data['name'] = name
            
        
            if (name_localizations is not ...):
                name_localizations = serializable_localized_dictionary_builder(name_localizations, 'name_localizations')
                if name_localizations is None:
                    name_localizations = {}
            
                data['name_localizations'] = name_localizations
        
        else:
            if (
                (description is not ...) or
                (description_localizations is not ...) or
                (name is not ...) or
                (name_localizations is not ...)
            ):
                if (achievement is None):
                    achievement = await self.achievement_get(achievement_id)
                
                if (name is not ...) or (name_localizations is not ...):
                    if (name is ...):
                        name = achievement.name
                    
                    if (name_localizations is ...):
                        name_localizations = achievement.name_localizations
                        if name_localizations is None:
                            name_localizations = {}
                        else:
                            name_localizations = name_localizations.copy()
                    else:
                        name_localizations = serializable_localized_dictionary_builder(
                            name_localizations, 'name_localizations'
                        )
                        if name_localizations is None:
                            name_localizations = {}
                    
                    name_localizations['default'] = name
                    data['name'] = name_localizations
                
                if (description is not ...) or (description_localizations is not ...):
                    if (description is ...):
                        description = achievement.description
                    
                    if (description_localizations is ...):
                        description_localizations = achievement.description_localizations
                        if description_localizations is None:
                            description_localizations = {}
                        else:
                            description_localizations = description_localizations.copy()
                    
                    else:
                        description_localizations = serializable_localized_dictionary_builder(
                            description_localizations,
                            'description_localizations',
                        )
                        if description_localizations is None:
                            description_localizations = {}
                    
                    description_localizations['default'] = description
                    data['description'] = description_localizations
        
        
        data = await self.api.achievement_edit(self.application.id, achievement_id, data)
        if achievement is None:
            achievement = Achievement(data)
        else:
            achievement._update_attributes(data)
        
        return achievement
    
    
    async def achievement_delete(self, achievement):
        """
        Deletes the passed achievement.
        
        This method is a coroutine.
        
        Parameters
        ----------
        achievement : ``Achievement``, `int`
            The achievement to delete.
        
        Raises
        ------
        TypeError
            If `achievement` was not given neither as ``Achievement``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        achievement_id = get_achievement_id(achievement)
        
        await self.api.achievement_delete(self.application.id, achievement_id)
    
    
    async def user_achievement_get_all(self, access):
        """
        Requests the achievements of a user with it's oauth2 access.
        
        This method is a coroutine.
        
        Parameters
        ----------
        access : ``Oauth2Access``, ``Oauth2User``, `str`.
            The access of the user, who's achievements will be requested.
        
        Returns
        -------
        achievements : `list` of ``Achievement`` objects
        
        Raises
        ------
        TypeError
            If `access` was not given neither as ``Oauth2Access``, ``Oauth2User``  or `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint is unintentionally documented and will never work. For reference:
        ``https://github.com/discordapp/discord-api-docs/issues/1230``.
        
        Always drops `DiscordException UNAUTHORIZED (401): 401: Unauthorized`.
        """
        if isinstance(access, (Oauth2Access, Oauth2User)):
            access_token = access.access_token
        elif isinstance(access, str):
            access_token = access
        else:
            raise TypeError(
                f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`, got '
                f'{access.__class__.__name__}; {access!r}.'
            )
        
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[AUTHORIZATION] = f'Bearer {access_token}'
        
        data = await self.api.user_achievement_get_all(self.application.id, headers)
        return [Achievement(achievement_data) for achievement_data in data]
    
    
    async def user_achievement_update(self, user, achievement, percent_complete):
        """
        Updates the `user`'s achievement with the given percentage. The  achievement should be `secure`. This
        method only updates the achievement's percentage.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ``ClientUserBase``, `int`
            The user, who's achievement will be updated.
        
        achievement : ``Achievement``, `int`
            The achievement, what's state will be updated
        
        percent_complete : `int`
            The completion percentage of the achievement.
            
            > Can be in range [0:100]. If outside is adjusted to the respective limit.
            
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase`` nor `int`.
            - If `achievement` was not given neither as ``Achievement``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint cannot grant achievement, but can it even update them?. For reference:
        ``https://github.com/discordapp/discord-api-docs/issues/1230``.
        
        Only secure updates are supported, if they are even.
        - When updating secure achievement: `DiscordException NOT FOUND (404), code = 10029: Unknown Entitlement`
        - When updating non secure: `DiscordException FORBIDDEN (403), code = 40001: Unauthorized`
        """
        user_id = get_user_id(user)
        
        achievement_id = get_achievement_id(achievement)
        
        assert _assert__user_achievement_update__percent_complete(percent_complete)
        
        if percent_complete < 0:
            percent_complete = 0
        elif percent_complete > 100:
            percent_complete = 100
        
        data = {'percent_complete': percent_complete}
        await self.api.user_achievement_update(user_id, self.application.id, achievement_id, data)
